@@ -269,7 +269,7 @@ gen_type_constraints(BTenv &env, C_Type* t, BTcause *cause)
       gen_type_constraints(env, *i, cause);
       // 2. bitfields must be dynamic if their widths are
       if ( m->hasValue() )
-        env.map[*m]->influence(env.map[*i],cause);
+        env.map[m->value()->type]->influence(env.map[*i],cause);
       // 3. if any field is dynamic, mark the struct as having a
       //    dynamic part
       env.add_dep_constraint(i->array_contents_type(),t,cause);
