@@ -528,13 +528,6 @@ TypeEnv::TypeEnv(CodeEnv &c)
   make_anno_links = true ;
 }
 
-C_Type*
-EllipsisType::trans(TypeEnv const&)
-{
-  DONT_CALL_THIS
-    return NULL;
-}
-
 C_Type* AbsType::trans(TypeEnv const &env) {
   BaseTypeTag mytag ;
   if ( strchr(properties,'i') )
@@ -602,18 +595,11 @@ C_Type*
 FunType::trans(TypeEnv const&env)
 {
   assert(cv.subsetof(constvol())); // functions are not const or volatile
-  // Tanslate the parameter types (keeping an eye on ellipsis).
-  bool has_var_args = 0;
+  // Tanslate the parameter types
   Plist<C_Type>* pars = new Plist<C_Type>;
-  foreach(t, *params, Plist<Type>) {
-    if ((*t)->isEllipsis) {
-      has_var_args = 1;
-      assert(++t == params->end());  // Ellipsis should be the last argument.
-      break;
-    }
+  foreach(t, *params, Plist<Type>)
     pars->push_back((*t)->trans(env));
-  }
-  C_Type *r = new C_Type(pars, ret->trans(env), has_var_args);
+  C_Type *r = new C_Type(pars, ret->trans(env), varargs);
   if ( env.make_anno_links )
     btannos.push_back(r);
   return r ;
@@ -1108,6 +1094,7 @@ NullExpr::trans(CodeEnv &env, ExprTrans &r,bool)
 void
 VarExpr::trans(CodeEnv &env, ExprTrans &r,bool)
 {
+  assert( isChecked() );
   assert( *type() == *decl->type );
   ObjectDecl *rdecl = decl ;
   while ( rdecl->refer_to != NULL ) rdecl = rdecl->refer_to ;
@@ -1160,6 +1147,7 @@ VarExpr::trans(CodeEnv &env, ExprTrans &r,bool)
 void
 CallExpr::trans(CodeEnv &env,ExprTrans &r,bool)
 {
+  assert( isChecked() );
   // Transform the components of the call.
   Interval precode ;
   Plist<C_Expr> *arglist = new Plist<C_Expr>;
@@ -1836,7 +1824,7 @@ SwitchStmt::trans(CodeEnv &env)
 
   if ( env2a.Default == NULL )
     env2a.Default = end ;
-  env2a.TestChain + env2a.Default ;
+  env2a.TestChain += env2a.Default ;
   return Interval(env2a.TestChain.begin,end);
 }
 
@@ -2009,32 +1997,12 @@ EnumDecl::trans(C_Pgm &pgm,TypeEnv const&env) {
   pgm.enumtypes.push_back(translated);
 }
 
-struct UsertypeTransCallback : StmtCallbackClosure {
-  C_Pgm &trans ;
-  TypeEnv const&env ;
-  UsertypeTransCallback(C_Pgm &t,TypeEnv const&e): trans(t), env(e) {}
-  void trans_these(Plist<UserDecl> &l) {
-    foreach(i,l,Plist<UserDecl>)
-      i->trans(trans,env);
-  }
-  void whatever(Stmt *s) {
-    CompoundStmt *c = s->this_as_compound();
-    if ( c != NULL )
-      trans_these(*c->usertypes);
-  }
-};
-
 static void
 trans_usertypes(CProgram &orig,C_Pgm &trans,TypeEnv const&env)
 {
-  UsertypeTransCallback callback(trans,env) ;
-  callback.trans_these(*orig.usertypes);
-  foreach(i,*orig.functions,Plist<FunDef>) {
-    if ( i->userdecls != NULL )
-      callback.trans_these(*i->userdecls) ;
-    if ( i->stmts != NULL )
-      i->stmts->preorder(&callback);
-  }
+  foreach(i,*orig.usertypes,Plist<UserDecl>)
+    if( i->refer_to == NULL)
+      i->trans(trans,env);
 }
 
 /************************************************************************
@@ -2053,9 +2021,17 @@ trans_globals1(CProgram &orig,C_Pgm &trans,TypeEnv const&env)
       continue ; // do not translate external functions
     assert(i->refer_to == NULL);
     assert(i->translated == NULL);
-    i->translated = new C_Decl(FunDf,i->type->trans(env),i->name);
+
+    C_Type *funtype = i->type->trans(env) ;
+    i->translated = new C_Decl(FunDf,funtype,i->name);
     if ( debug ) debug << "made fun frame: " << (void*)i->translated
                        << " for " << i->name << endl ;
+
+    // translate parameters
+    Plist<C_Type>::iterator ti = funtype->fun_params().begin();
+    foreach(pi,*i->decls,Plist<VarDecl>)
+      pi->translated = new C_Decl(VarDcl,*ti++,pi->name) ;
+    assert(!ti);
   }
   foreach(ii,*orig.definitions,Plist<VarDecl>) {
     if ( ii->refer_to != NULL )
@@ -2167,14 +2143,11 @@ FunDef::trans(CodeEnv &env)
   Plist<C_Decl> *paramlist = new Plist<C_Decl>;
     
   // translate parameters
-  Plist<C_Type>::iterator ti = f->type->fun_params().begin();
   foreach(pi,*decls,Plist<VarDecl>) {
-    pi->translated = new C_Decl(VarDcl,*ti++,pi->name) ;
     Interval I = pi->trans(theenv) ;
     assert( I.trivial() );
     paramlist->push_back(pi->translated);
   }
-  assert( ti == f->type->fun_params().end() );
     
   Interval code = stmts->trans(theenv) ;
   code += theenv.returnTarget ; // now code cannot be trivial
