@@ -8,6 +8,10 @@
  */
 
 #include <cmixconf.h>
+#ifdef HAVE_REGEX_H
+#include <stddef.h> // for redhat machines which otherwise choke on regex.h
+#include <regex.h>  // for OSF alphas with buggy GCC installations
+#endif
 #include <ctype.h>
 #include <string.h>
 #include "auxilary.h"
@@ -208,21 +212,6 @@ decaycheck(Expr *&exp)
   return t ;
 }
 
-static Type*
-decaychecktype(Type*& typ)
-{
-  Type* t = typ;
-  // "Array to T" decays to "pointer to T"
-  if(t->realtype==ArrayT) {
-    typ = new PtrType(((ArrayType*)t)->next);
-  }
-  // Functions decay to pointers to themselves
-  else if(t->realtype==FunT) {
-    typ = new PtrType(t);
-  }
-  return typ;
-}
-
 //////////////////////////////////////
 // CHECKING INITIALIZERS
 //
@@ -283,7 +272,6 @@ split_init_list(Plist<Init>::iterator &i,Type *t, bool master)
 
     switch ( t->realtype ) {
     case FunT:
-    case EllipsisT:
     case BaseT:
     case PtrT:
     case EnumT:
@@ -466,68 +454,37 @@ Type* currentFunRetType = NULL;
 Type*
 FunDef::check()
 {
-  Type* savedCurrentFunRetType;
-  
   if (isChecked()) return type;
   // Check linkage.
   assert(linkage==External || linkage==Internal);
-
-  // Check return type; it must not be an array.
   assert(type->isFun);
-  Type* rt = ((FunType*)type)->ret;
-  savedCurrentFunRetType = currentFunRetType; // Save previous value
-  currentFunRetType = rt; // Global variable used in ReturnStmt::check().
-
-  // Let parameters decay.
-  for(Plist<Type>::mutable_iterator par = ((FunType*)type)->params->begin() ;
-      par ; par++ ) {
-    Type *parameter = *par ;
-    decaychecktype(parameter);
-    par << parameter ;
-  }
+  type->check() ;
 
   // If this is a function prototype it has to be external.
   if (stmts==NULL) {
+    assert(decls==NULL);
+    type->check() ;
     if (linkage!=External)
       Diagnostic(ERROR,pos) << name << "() never gets a body" ;
   } else {
-    // Run through the user types.
-    if (userdecls==NULL)
-      userdecls = new Plist<UserDecl>();
-    else {
-      foreach(ud, *userdecls, Plist<UserDecl>) (*ud)->check();
-    }
-    // Run through the parameters.
+    FunType *funt = (FunType*)type ;
+    #ifndef NDEBUG
+    Plist<Type>::iterator type = funt->params->begin() ;
     foreach(var, *decls, Plist<VarDecl>) {
-      // Turn functions and arrays into pointers.
-      decaychecktype((*var)->type);
+      assert(var->type == *type);
+      type++ ;
       assert((*var)->linkage!=Typedef);
       assert((*var)->linkage!=AbstractType);
       assert((*var)->tag==Vardecl);
-      if ((*var)->type==NULL)
-        Diagnostic(INTERNAL,(*var)->pos) << "parameter out of order";
-      if ((*var)->init!=NULL)
-        Diagnostic(ERROR,(*var)->pos) << "parameters cannot have initializers";
+      assert(var->init==NULL);
     }
+    assert( !type );
+    #endif
+    assert(currentFunRetType==NULL);
+    // Global variable used in ReturnStmt::check().
+    currentFunRetType = funt->ret;
     stmts->check();
-  }
-  currentFunRetType = savedCurrentFunRetType; // Restore previous value
-  switch(rt->realtype) {
-  case BaseT:
-  case PtrT:
-  case StructT:
-  case UnionT:
-  case EnumT:
-    break ;
-  case AbstractT:
-    if ( strchr(((AbsType*)rt)->properties,'p') || rt->isArithmic() )
-      break ;
-    // else fall through
-  case ArrayT:
-  case FunT:
-  case EllipsisT:
-    Diagnostic(ERROR,pos) << "a function cannot return this type";
-    break ;
+    currentFunRetType = NULL ;
   }
   return type;
 }
@@ -586,7 +543,7 @@ bool BaseType::isObject() {
 void BaseType::check()
 {
   isChecked();
-  assert(!(isFun || isEllipsis));  
+  assert(!isFun);  
   // void is not a complete type.
 }
 
@@ -598,16 +555,6 @@ void AbsType::check() {
     isChecked();
 }
 
-bool EllipsisType::isObject() {
-    return false ;
-}
-
-void EllipsisType::check()
-{
-  isChecked();
-  assert(!(cv.cst || cv.vol || isFun || isVoid || !isEllipsis));  
-}
-
 bool FunType::isObject() {
     return false ;
 }
@@ -615,10 +562,27 @@ bool FunType::isObject() {
 void FunType::check()
 {
   if (isChecked()) return;
-  // Dont run through the parameter types.
-  // check the return type;
-  assert(!(!isFun || isVoid || isEllipsis || cv.vol || cv.cst));
+  assert(!(!isFun || isVoid || cv.vol || cv.cst));
+
+  foreach(i,*params,Plist<Type>)
+    i->check();
   ret->check() ;
+  switch(ret->realtype) {
+  case BaseT:
+  case PtrT:
+  case StructT:
+  case UnionT:
+  case EnumT:
+    break ;
+  case AbstractT:
+    if ( strchr(((AbsType*)ret)->properties,'p') || ret->isArithmic() )
+      break ;
+    // else fall through
+  case ArrayT:
+  case FunT:
+    Diagnostic(ERROR,pos) << "function returning " << ret ;
+    break ;
+  }
 }
 
 bool ArrayType::isObject() {
@@ -628,14 +592,12 @@ bool ArrayType::isObject() {
 void ArrayType::check()
 {
   if (isChecked()) return;
-  assert(!(isFun & isVoid & isEllipsis));
+  assert(!(isFun || isVoid));
   next->check();
   if (size != NULL) {
-    if (!decaycheck(size)) {
-      Diagnostic(WARNING,size->pos)
-        << "something is wrong with array size expression";
-      return;
-    }
+    Type *stype = decaycheck(size);
+    if( !stype->isIntegral() )
+      Diagnostic(ERROR,size->pos) << "non-integral array size" ;
   }
 }
 
@@ -646,7 +608,7 @@ bool PtrType::isObject() {
 void PtrType::check()
 {
   if (isChecked()) return ;
-  assert(!(isFun || isVoid || isEllipsis));
+  assert(!(isFun || isVoid));
   next->check();
 }
 
@@ -657,7 +619,7 @@ bool EnumType::isObject() {
 void EnumType::check()
 {
   if (isChecked()) return ;
-  assert(!(isFun || isVoid || isEllipsis));  
+  assert(!(isFun || isVoid));  
   def->check() ;
 }
 
@@ -668,7 +630,7 @@ bool UnionType::isObject() {
 void UnionType::check()
 {
   if (isChecked()) return;
-  assert(!(isFun || isVoid || isEllipsis));  
+  assert(!(isFun || isVoid));  
   def->check() ;
 }
 
@@ -679,7 +641,7 @@ bool StructType::isObject() {
 void StructType::check()
 {
   if (isChecked()) return;
-  assert(!(isFun || isVoid || isEllipsis));  
+  assert(!(isFun || isVoid));  
   def->check();
 }
 
@@ -748,24 +710,10 @@ DefaultStmt::check()
 void
 CompoundStmt::check()
 {
-  Plist<UserDecl>::iterator firstUO = usertypes->begin();
-  Plist<UserDecl>::iterator lastUO  = usertypes->end();
-  while (firstUO != lastUO) {
-    (*firstUO)->check();
-    firstUO++;
-  }
-  Plist<VarDecl>::iterator firstOb = objects->begin();
-  Plist<VarDecl>::iterator lastOb  = objects->end();
-  while (firstOb != lastOb) {
-    (*firstOb)->check();
-    firstOb++;
-  }
-  Plist<Stmt>::iterator firstS = stmts->begin();
-  Plist<Stmt>::iterator lastS  = stmts->end();
-  while (firstS != lastS) {
-    (*firstS)->check();
-    firstS++;
-  }
+  foreach(i,*objects,Plist<VarDecl>)
+    i->check();
+  foreach(i,*stmts,Plist<Stmt>)
+    i->check();
 }
 
 void
@@ -842,13 +790,11 @@ Indirection::check()
 {
   if(stmt == NULL)
     Diagnostic(ERROR,pos) << "label not found in goto statement";
-  stmt->check();
 }
 
 void
 GotoStmt::check()
 {
-  if (isChecked()) return;
   assert(ind != NULL);
   ind->check();
 }
@@ -979,7 +925,7 @@ VarExpr::check()
 {
   if (isChecked()) return exprType;
   assert(decl != NULL);
-  Type* t = decl->check();
+  Type* t = decl->type;
   setRealType(t);
   // enumeration and magic constants are rvals
   if ( decl->tag == Vardecl )
@@ -1017,16 +963,11 @@ CallExpr::check()
     // Check arguments against the formal parameters.
     Plist<Type>::iterator firstPar = funt->params->begin() ;
     Plist<Expr>::mutable_iterator firstArg = args->begin();
-    int nargs = 0;
-    while (firstArg && firstPar) {
+    while (firstArg) {
       Expr *argument = *firstArg ;
-      nargs++;
-      if (debug^5) debug << '(' << nargs << ')';
       assert(argument!=NULL);
       decaycheck(argument);
-      // If ellipsis is the formal parameter, skip the test.
-      assert(firstPar);
-      if (!(*firstPar)->isEllipsis) {
+      if( firstPar ) {
         // Check that types agree, convert if necessary.
         checked_force((*firstPar),argument);
         firstPar++;
@@ -1034,13 +975,14 @@ CallExpr::check()
       firstArg << argument ;
       firstArg++;
     }
-    if (firstArg)
+    unsigned nparms = funt->params->size();
+    if ( !funt->varargs && args->size() > nparms )
       Diagnostic(ERROR,pos) << "too many arguments in function "
-                            << "call (expecting " << nargs << ')' ;
+                            << "call (expecting " << nparms << ')' ;
     // Check that the function does not expect more arguments.
-    if ( firstPar && !firstPar->isEllipsis)
-      Diagnostic(ERROR,pos) << "expecting more than " << nargs
-                            << " arguments in function call" ;
+    if ( firstPar )
+      Diagnostic(ERROR,pos) << "function call needs " << nparms
+                            << " arguments" ;
     setRealType(funt->ret);
   }
   else {
