@@ -62,12 +62,12 @@ makeConstExpr(long i, const Position p)
 
 Type::Type(TypeTag t, bool fun, bool ell)
   : checked(0), realtype(t), cv(false,false),
-    isFun(fun), isEllipsis(ell), isVoid(false)
+    isFun(fun), isVoid(false)
 {}
 
 Type::Type(Type& t)
   : checked(0), realtype(t.realtype), cv(t.cv),
-    isFun(t.isFun), isEllipsis(t.isEllipsis), isVoid(t.isVoid)
+    isFun(t.isFun), isVoid(t.isVoid)
 {}
 
 bool Type::operator==(Type &other) {
@@ -101,19 +101,6 @@ BaseType::show(ostream& os)
   os << basetype2str(tag);
 }
 
-bool EllipsisType::equals(Type* other) {
-    assert(other != NULL);
-    if (other->realtype != realtype) return false ;
-    return true ;
-}
-
-void
-EllipsisType::show(ostream& os)
-{
-  Type::show(os);
-  os << "...";
-}
-
 bool AbsType::equals(Type *other) {
     if ( other->realtype != AbstractT )
         return 0 ;
@@ -160,7 +147,7 @@ ArrayType::show(ostream& os)
 bool
 FunType::unspecified()
 {
-  return params->size() == 1 && params->front()->realtype == EllipsisT ;
+  return params->empty() && varargs ;
 }
 
 bool FunType::equals(Type* other) {
@@ -174,12 +161,8 @@ bool FunType::equals(Type* other) {
     // Check the argument types by running through the list.
     Plist<Type>::iterator it1 = params->begin();
     Plist<Type>::iterator it2 = real->params->begin();
-    while (it1 != params->end() && it2 != real->params->end())
+    while (it1 && it2)
     {
-        // We assume that ellipsis only can be the last type in the list.
-        // XXX this is too conservative (but who expects C-Mix to find
-        // errors as subtle as this? GeGen might do one day
-        if ((*it1)->isEllipsis || (*it2)->isEllipsis) return 1;
         // the qualifiers on parameter types matter only inside the function
         if (!( **it1 == **it2 ))
             return false ;
@@ -187,8 +170,9 @@ bool FunType::equals(Type* other) {
         if (debug) debug << "[parameter OK]";
     }
     // Check equal length.
-    if (it1 != params->end() || it2 != real->params->end()) return 0;
-    return 1;
+    if( it1 ) return real->varargs ;
+    if( it2 ) return varargs ;
+    return true;
 }
 
 void
@@ -677,12 +661,6 @@ AssignExpr::show(ostream& os)
 
 ///////// Deep Copy ///////////////
 
-EllipsisType*
-EllipsisType::copy()
-{
-    return new EllipsisType(*this);
-}
-
 AbsType* AbsType::copy() {
     return new AbsType(*this) ;
 }
@@ -738,13 +716,6 @@ EnumType::copy()
     return new EnumType(*this);
 }
 
-///////////////////////////////////
-
-bool EllipsisType::isIntegral()  { return false; }
-bool EllipsisType::isArithmic()  { return false; }
-bool EllipsisType::isAggregate() { return false; }
-bool EllipsisType::isDerived()   { return false; }
-
 //////////////////////////////////////////
 
 bool AbsType::isIntegral()  { return strchr(properties,'i'); }
@@ -931,7 +902,6 @@ CompoundStmt::show(ostream& os)
 {
   os << "{ ";
   foreach( ob, *objects, Plist<VarDecl> ) { ob->show(os); os << "; "; }
-  foreach( ud, *usertypes, Plist<UserDecl> ) { ud->show(os); os << "; "; }
   foreach( s, *stmts, Plist<Stmt> ) { s->show(os); os << "; "; }
   os << " }";
 }
