@@ -118,7 +118,6 @@ class FunDef : public ObjectDecl
 {
 public:
     Plist<VarDecl>* decls;
-    Plist<UserDecl>* userdecls;
     Stmt* stmts;
     StateAnnoAtom *effects ;
     CallAnnoAtom *calltime ;
@@ -126,10 +125,9 @@ public:
            Position p,
            Type* t,
            Linkage link,
-           Plist<VarDecl>* ds,
-           Plist<UserDecl>* us)
-        : ObjectDecl(n,p,t,Fundef,link),
-          decls(ds), userdecls(us), stmts(NULL), effects(NULL), calltime(NULL)
+           Plist<VarDecl>* ds)
+        : ObjectDecl(n,p,t,Fundef,link), decls(ds),
+          stmts(NULL), effects(NULL), calltime(NULL)
         {}
     OBJECTVIRTUALS(FunDef,;)
     void trans(CodeEnv &) ;
@@ -206,10 +204,9 @@ public:
 
 ////// TYPES //////
 
-typedef enum { EllipsisT, BaseT, PtrT, ArrayT, FunT,
+typedef enum { BaseT, PtrT, ArrayT, FunT,
                StructT, UnionT, EnumT, AbstractT } TypeTag;
 /*	switch (t1->realtype) {
-	case EllipsisT:
 	case BaseT:
 	case PtrT:
 	case ArrayT:
@@ -246,7 +243,6 @@ public:
     const TypeTag realtype; // Used to compare types.
     constvol cv ;
     const bool isFun; // is this a function type?
-    const bool isEllipsis;
     bool isVoid;
     Type(TypeTag t, bool fun=false, bool ell=false);
     Type(Type& t);
@@ -264,14 +260,6 @@ public:
     bool operator==(Type &);
 };
 
-class EllipsisType : public Type
-{
-public:
-    EllipsisType() : Type(EllipsisT, false,true) {}
-    EllipsisType(EllipsisType& t) : Type(t) {}
-    TYPEVIRTUALS(EllipsisType,;);
-};
-
 class AbsType : public Type {
 public:
     char const *name ;
@@ -317,14 +305,12 @@ class FunType : public Type {
   public:
     Plist<Type>* params;
     Type* ret;
-    FunType(FunType& t) : Type(t), params(t.params), ret(t.ret) {}
-    FunType(Plist<Type>* par, Type* t)
-        : Type(FunT, true), params(par), ret(t)
-    {
-        // If there is exactly one void parameter, remove it.
-        if (params->size() == 1 && params->front()->isVoid)
-            params->pop_front();
-    }
+    bool varargs ;
+    Position pos ;
+    FunType(FunType& t)
+      : Type(t), params(t.params), ret(t.ret), varargs(t.varargs) {}
+    FunType(Plist<Type>* par, bool v, Type* t, Position p)
+      : Type(FunT, true), params(par), ret(t), varargs(v), pos(p) {}
     bool unspecified() ;
     
     TYPEVIRTUALS(FunType,;);
@@ -391,7 +377,7 @@ class Stmt
 public:
     const Position pos;
     Stmt(Position p) : checked(0), pos(p) {}
-    bool isChecked() { return checked || ( checked = true , false ); }
+//    bool isChecked() { return checked || ( checked = true , false ); }
     STMTVIRTUALS(Stmt,=0;)
     virtual void preorder(StmtCallbackClosure*) =0 ;
     virtual CompoundStmt *this_as_compound();
@@ -443,13 +429,11 @@ class CompoundStmt : public Stmt
 {
 public:
     Plist<VarDecl>* objects;
-    Plist<UserDecl>* usertypes;
     Plist<Stmt>* stmts;
     CompoundStmt (Position p,
                   Plist<VarDecl>* os,
-                  Plist<UserDecl>* uts,
                   Plist<Stmt>* ss)
-        : Stmt(p), objects(os), usertypes(uts), stmts(ss) {}
+        : Stmt(p), objects(os), stmts(ss) {}
     STMTVIRTUALS(CompoundStmt,;)
     virtual void preorder(StmtCallbackClosure*) ;
     virtual CompoundStmt *this_as_compound();
