@@ -304,6 +304,30 @@ GETSET(Use)  // defines DFresult::getUseSet
 
 #undef GETSET
 
+Pset<C_Decl> const &
+DFresult::GetMemoSet(C_BasicBlock *bb) const
+{
+  InvALocSet const *set = Use[bb] ;
+  assert(!set->isInversed());
+  return set->theSet();
+}
+
+Pset<C_Decl> const &
+DFresult::GetReadSet(C_Decl *fun) const
+{
+  InvALocSet const *set = Use[fun] ;
+  assert(!set->isInversed());
+  return set->theSet();
+}
+
+Pset<C_Decl> const &
+DFresult::GetWriteSet(C_Decl *fun) const
+{
+  InvALocSet const *set = Wmay[fun] ;
+  assert(!set->isInversed());
+  return set->theSet();
+}
+
 
 // ===========================================================================
 // '''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
@@ -736,107 +760,6 @@ DFresult::Analyse(C_Pgm const& pgm, PAresult const& pa, ALocSet const& ones)
   if (dfdebug^1) dfdebug << endl;  
 }
 
-// ===========================================================================
-// '''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
-//  Translation of the results to strings (used by gegen).
-// ,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,
-// ===========================================================================
-
-static void
-outputSetAsString(Nset& IDset, unsigned IDmax, const char hit,
-		  const char miss, GegenStream& ost)
-{
-  // Traverse the translated set. 
-  unsigned id = 1;
-  foreach( n, IDset, Nset ) {
-    // Position 0 is reserved (not used).
-    if (*n == 0) continue;
-    while (id < *n) { ost << miss; id++; } 
-    ost << hit; id++;
-  }
-  assert(id <= IDmax+1);
-  while (id++ <= IDmax+1) ost << miss;
-}
-
-static C_Decl*
-outerMostContainer(C_Decl* d)
-{
-  while ( d->isContained() ) {
-    C_Decl* container = d->containedIn();
-    if ( container->type->tag == Function ) break;
-    d = container;
-  }
-  return d;
-}
-
-static void
-enumerate(InvALocSet const *set, Plist<C_Decl> &list)
-{
-  assert(set);
-  assert(!set->isInversed());
-  foreach(d, set->theSet(), Pset<C_Decl>)
-    list.push_back(*d);
-}
-
-static void
-setToString(Plist<C_Decl> const &set, IdNumberMap const& IDmap,
-            unsigned IDmax, GegenStream &ost)
-{
-  // This set holds the translated set.
-  Nset IDset;
-  if (dfdebug^1) dfdebug << "[Approximating in-use: ";
-  foreach( d, set, Plist<C_Decl> ) {
-    // Approximate the result: if some member of a struct/union/array is
-    // in-use, the whole containing object is in-use.
-    C_Decl* decl = outerMostContainer(*d);
-    if (dfdebug^1) if ( decl != *d ) 
-      dfdebug << d->get_name() << str_rarrow << decl->get_name();      
-    unsigned id = IDmap[*d];
-    IDset.insert(id);
-  }  
-  if (dfdebug^1) dfdebug << str_rbracket;
-  if (dfdebug^1) {
-    // Write out the set.
-    ost << "/* { ";
-    foreach( n, IDset, Nset ) {
-      ost << *n << ' ';
-    }
-    ost << "} */";
-  }
-  ost << "\"";
-  outputSetAsString(IDset, IDmax, '1', '0', ost);
-  ost << "\"";
-}
-  
-void
-DFresult::EmitMemoParam(C_BasicBlock *bb,
-                        IdNumberMap const& IDmap,
-                        unsigned IDmax,
-                        GegenStream &ost) const
-{
-  // Transform the set into a string of 0s and 1s ("1" means "in-use").
-
-  Plist<C_Decl> interset ;
-  // Get the use-set.
-  enumerate(Use[bb],interset);
-  setToString(interset,IDmap,IDmax,ost);
-}
-
-void
-DFresult::EmitMemoParam(C_Decl *fun,
-                        IdNumberMap const& IDmap,
-                        unsigned IDmax,
-                        GegenStream &ost) const
-{
-  // Transform the set into a string of 0s and 1s ("1" means "must-memoize").
-
-  Plist<C_Decl> interset ;
-  // Get the use-set and the may-write-set
-  enumerate(Use[fun],interset);
-  enumerate(Wmay[fun],interset);
-  setToString(interset,IDmap,IDmax,ost);
-}
-
 /////////////////////////////////////////////////////////////////////////////
 
 OutDf::OutDf(DFresult const&d, CoreAnnotator *nxt)
