@@ -36,10 +36,13 @@ typedef union{
 #define	CHARS	284
 #define	ARE	285
 #define	GLYPHS	286
-#define	WORD	287
-#define	INT	288
-#define	IDENT	289
-#define	STRING	290
+#define	LIFT	287
+#define	LONG	288
+#define	DOUBLE	289
+#define	WORD	290
+#define	INT	291
+#define	IDENT	292
+#define	STRING	293
 
 
 extern YYSTYPE cmxlval;
