@@ -76,27 +76,26 @@ struct FunGegen {
   bool use_cmixSR ;
   bool cmixSRisExp ;
   bool use_pendinglist ;
-  char const *sharestrategy ;
   array<C_BasicBlock,unsigned> speclabels ;
     
   FunGegen(GegenEnv*,C_Decl *);
 
   void EmitStatement(C_Stmt *s) const;
-  bool EmitGoto(C_BasicBlock *, bool emitit) const;
+  virtual bool EmitGoto(C_BasicBlock *, bool emitit) const;
   bool EmitJump(C_Jump *j) const;
   void EmitReturnSequence(C_Expr *e) const ;
   void EmitBasicBlock(C_BasicBlock *) const;
   void ConstructResidualCall() const;
+
+  virtual ~FunGegen() {}
 };
 
 FunGegen::FunGegen(GegenEnv *e, C_Decl *f)
   : env(*e), fun(f), returntype(f->fun_ret()), speclabels(0)
 {
-  // void is never dynamic, and *if* it should ever be, the
-  // calling sequence expects it to be treated as dynamic.
-  if ( env.bt.Dynamic(returntype) ) retmode = DYNAMIC ;
-  else if ( returntype->isVoid() )  retmode = VOID ;
-  else                              retmode = STATIC ;
+  if ( returntype->isVoid() )            retmode = VOID ;
+  else if ( env.bt.Dynamic(returntype) ) retmode = DYNAMIC ;
+  else                                   retmode = STATIC ;
 }
 
 void
@@ -145,7 +144,7 @@ FunGegen::EmitStatement(C_Stmt *s) const
         } else
           env.ost << userhole << exit << ',' ;
         env.ost << exit ; // almost no-op: the userhole was last
-      } else if ( env.bt.Dynamic(ret) )
+      } else if ( env.bt.Dynamic(ret) && !ret->isVoid() )
         env.ost << "cmixStmt(\"?\",", nparens++ ;
       env.ost << EmitExpr(s->call_expr(),30) << '(' , nparens++ ;
       bool first = true ;
@@ -264,18 +263,8 @@ FunGegen::EmitStatement(C_Stmt *s) const
 bool
 FunGegen::EmitGoto(C_BasicBlock *bb, bool emitit) const
 {
-  unsigned speclabel = speclabels[bb] ;
-  if ( speclabel != 0 ) {
-    if ( emitit ) env.ost << "cmixGoto(" ;
-    env.memo.emit_pendinsert(fun,bb,speclabel);
-    if ( emitit ) env.ost << ");\n" ;
+  DONT_CALL_THIS
     return false ;
-  } else {
-    env.ost << "goto cmixL" << bb->Numbered_ID ;
-    if ( emitit )
-      env.ost << ";\n" ;
-    return true ;
-  }
 }
 
 static void
@@ -344,7 +333,7 @@ FunGegen::EmitJump(C_Jump *s) const
           
       env.ost << "  cmixIf(" ;
       EmitGoto(s->cond_then(),false);
-      env.ost << ",\n\t"  ;
+      env.ost << ','  ;
       EmitGoto(s->cond_else(),false);
       env.ost << ",\n\t" << enter ;
       if ( dlist.size() == 1 )
@@ -408,7 +397,9 @@ FunGegen::EmitReturnSequence(C_Expr *e) const
 {
   env.ost << "  cmixPopLocals(" << DOcounter << ");\n";
   if ( !unfolding ) {
-    env.ost << "  cmixPopFun(cmix_params, " << sharestrategy << ");\n" ;
+    if ( !env.bt.Dynamic(fun) )
+      env.memo.emit_endstate(fun) ;
+    env.ost << "  cmixPopFun();\n" ;
     if ( retmode == DYNAMIC ) {
       env.ost << "  return cmixMkExp(" ;
       ConstructResidualCall();
@@ -440,7 +431,7 @@ FunGegen::EmitReturnSequence(C_Expr *e) const
 void
 FunGegen::EmitBasicBlock(C_BasicBlock *bb) const
 {
-  if ( use_assigned_gotos || speclabels[bb] == 0 )
+  if ( speclabels[bb] == 0 )
     env.ost << "cmixL" << bb->Numbered_ID << ":\n" ;
   else
     env.ost << "case " << speclabels[bb] << ": /* cmixL"
@@ -477,13 +468,28 @@ FunGegen::ConstructResidualCall() const
 /////////////////////////////////////////////////////////////////////////////
 /////////////////////////////////////////////////////////////////////////////
 
-struct StdFunGegen : FunGegen {
+class StdFunGegen : FunGegen {
+  GegenLocalMemo localmemo ;
+  Plist<C_BasicBlock> pendusers ;
+  void penduser(C_BasicBlock *);
+  
+  virtual bool EmitGoto(C_BasicBlock *, bool emitit) const;
+public:
   StdFunGegen(GegenEnv*,C_Decl *);
   void emit() ;
+  virtual ~StdFunGegen() {}
 };
 
+void
+StdFunGegen::penduser(C_BasicBlock *bb) {
+  if ( speclabels[bb] == 0 ) {
+    pendusers.push_back(bb) ;
+    speclabels[bb] = pendusers.size() ;
+  }
+}
+  
 StdFunGegen::StdFunGegen(GegenEnv *e, C_Decl *f)
-  : FunGegen(e,f)
+  : FunGegen(e,f), localmemo(e->memo,f)
 {
   // we unfold exactly the functions that are marked
   // as unsharable.
@@ -493,26 +499,23 @@ StdFunGegen::StdFunGegen(GegenEnv *e, C_Decl *f)
   // conditional
   // AND (makholm 1999-07-14) any BB that ends with a dynamic conditional,
   // unless it is not itself under dynamic control.
-  unsigned speclabelcounter = 0 ;
   foreach(bb,fun->blocks(),Plist<C_BasicBlock>) {
     C_Jump *j = bb->exit() ;
     if ( j->tag == C_If && env.bt.Dynamic(j->cond_expr()->type) ) {
-      if ( env.bt.Dynamic(*bb) && speclabels[*bb]==0 )
-        speclabels[*bb] = ++speclabelcounter ;
-      if ( speclabels[j->cond_then()]==0 )
-        speclabels[j->cond_then()] = ++speclabelcounter ;
-      if ( speclabels[j->cond_else()]==0 )
-        speclabels[j->cond_else()] = ++speclabelcounter ;
+      if ( env.bt.Dynamic(*bb) )
+        penduser(*bb);
+      penduser(j->cond_then());
+      penduser(j->cond_else());
     }
   }
-  use_pendinglist = speclabelcounter > 0 ;
+  use_pendinglist = !pendusers.empty();
     
     // if memo_unconditional_mode is set, use pending list for ALL
     // BBs that are under dynamic control
   if ( use_pendinglist && memo_unconditional_mode )
     foreach(bb,fun->blocks(),Plist<C_BasicBlock>)
-      if ( env.bt.Dynamic(*bb) && speclabels[*bb]==0 )
-        speclabels[*bb] = ++speclabelcounter ;
+      if ( env.bt.Dynamic(*bb) )
+        penduser(*bb);
     
     // Bugger. With gcc 2.8.1 on the HPs, P_gen tends to be
     // incorrect and get a SIGSEGV when the return is in
@@ -526,13 +529,23 @@ StdFunGegen::StdFunGegen(GegenEnv *e, C_Decl *f)
   use_cmixSR = retmode==STATIC || retmode==DYNAMIC && unfolding ;
 #endif
   cmixSRisExp = use_cmixSR && retmode==DYNAMIC && !use_pendinglist ;
+}
 
-  if ( env.sa.find(fun) )
-    sharestrategy = "cmixNonshared";
-  else if ( env.bt.Dynamic(fun) )
-    sharestrategy = "cmixSharedPure";
-  else
-    sharestrategy = "cmixSharedSideEffecting";
+bool
+StdFunGegen::EmitGoto(C_BasicBlock *bb, bool emitit) const
+{
+  unsigned speclabel = speclabels[bb] ;
+  if ( speclabel != 0 ) {
+    if ( emitit ) env.ost << "cmixGoto(" ;
+    env.ost << "cmixPendinsert(" << speclabel << ")" ;
+    if ( emitit ) env.ost << ");\n" ;
+    return false ;
+  } else {
+    env.ost << "goto cmixL" << bb->Numbered_ID ;
+    if ( emitit )
+      env.ost << ";\n" ;
+    return true ;
+  }
 }
 
 void
@@ -543,10 +556,13 @@ StdFunGegen::emit()
   env.specializable_heading(fun,false) ;
   env.ost << "{\n" ;
 
-  if ( !unfolding )
-    env.ost << "  Code cmix_fun;\n" ;
+  if ( !unfolding ) {
+    env.ost <<
+      "  Code cmix_fun;\n"
+      "  static struct cmixFunPool cmix_funpool = {0};\n" ;
+  }
     
-    // local variables
+  // local variables
   foreach(j,fun->fun_locals(),Plist<C_Decl>) {
     env.ost << "  " << *j ;
     if ( env.spectime(*j) && j->hasInit() )
@@ -558,7 +574,7 @@ StdFunGegen::emit()
     if ( !env.spectime(*jj) )
       env.ost << "  " << *jj << ";\n" ;
     
-    // static return value placeholders
+  // static return value placeholders
   if ( use_cmixSR ) {
     if ( retmode == STATIC )
       env.ost << "  " << Unqualified(returntype,"cmixSR") ;
@@ -575,8 +591,19 @@ StdFunGegen::emit()
   if ( unfolding && use_pendinglist )
     env.ost << "  cmixLabelTag cmixEndLabel = cmixMakeLabel();\n" ;
 
-  DOcounter = env.memo.emit_localmemodata(fun,!unfolding,use_pendinglist);
+  DOcounter = localmemo.emit_data(!unfolding,use_pendinglist);
 
+  if ( use_pendinglist ) {
+    env.ost << "  cmixLocalMemoWhat cmix_blocks[] = { 0" ;
+    foreach(bb,pendusers,Plist<C_BasicBlock>) {
+      env.ost << ",\n    /* cmixL" << bb->Numbered_ID << " */ " ;
+      localmemo.emit_memodata(*bb);
+    }
+    env.ost << "};\n" ;
+  }
+  
+  localmemo.emit_code();
+  
   if ( use_cmixSR && retmode == DYNAMIC && !cmixSRisExp )
     env.ost << "  cmixDeclare(cmixLocal,cmixSR," << enter
             << pushtype(returntype,constvol()) << userhole
@@ -595,9 +622,8 @@ StdFunGegen::emit()
     }
   } else {
     // not unfolding: emit function sharing and creation code
-    env.ost << "  cmix_fun = cmixFunctionSeenB4(" ;
     env.memo.emit_funmemo(fun);
-    env.ost << ");\n"
+    env.ost << 
       "  if (cmix_fun) {\n"
       "    cmixPopLocals(" << DOcounter << ");\n" ;
     if ( retmode == DYNAMIC ) {
@@ -619,9 +645,7 @@ StdFunGegen::emit()
 
         // now we can create the residual function itself
     env.ost << "  cmix_fun = " << env.resnames.request(fun->get_name())
-            << ";\n  cmixPushFun(";
-    env.memo.emit_funmemo(fun);
-    env.ost << ",\n\tcmix_fun," << sharestrategy << ",\n\t" ;
+            << ";\n  cmixFunHeading(cmix_fun," ;
     residual_heading(env,fun,"static ","cmix_fun");
     env.ost << ");\n" ;
   }
@@ -645,26 +669,16 @@ StdFunGegen::emit()
     // pending loop control:
       
     // create a new pending list
-    if ( use_assigned_gotos )
-      env.ost << "  cmixPushPend_v();\n  " ;
-    else
-      env.ost << "  cmixPushPend();\n  " ;
+    env.ost << "  cmixPushPend(cmix_locals,cmix_blocks,"
+            << pendusers.size() << ");\n  " ;
 
     // if the first BB is not handled by the pending list, simply
     // jump to it. Else insert it properly into the pending list
     EmitGoto(fun->blocks().front(),false);
-    env.ost << ";\n" ;
-    if ( use_assigned_gotos ) {
-      env.ost << "cmix_loop: {\n"
-        "    void *cmix_sp = cmixPending_v();\n"
-        "    if ( cmix_sp ) goto *cmix_sp ;\n"
-        "  }\n"
-        "  cmixPopPend_v();\n";
-    } else {
-      env.ost << "cmix_loop: switch(cmixPending()) {\n"
-        "  case 0:\n"
-        "  cmixPopPend();\n" ;
-    }
+    env.ost << ";\n" 
+      "cmix_loop: switch(cmixPending()) {\n"
+      "  case 0:\n"
+      "  cmixPopPend();\n" ;
     if ( unfolding )
       env.ost << "  cmixLabel(cmixEndLabel);\n" ;
     // GCC 2.8.1 has been known to fail on early returns...
@@ -681,7 +695,7 @@ StdFunGegen::emit()
   foreach(bb,fun->blocks(),Plist<C_BasicBlock>)
     EmitBasicBlock(*bb);
 
-  if ( use_pendinglist && !use_assigned_gotos )
+  if ( use_pendinglist )
     env.ost << "  } /* end of pending blocks */\n" ;
     
   env.ost << "cmix_return:\n" ;
@@ -695,6 +709,7 @@ StdFunGegen::emit()
 struct EntryPointGegen : FunGegen {
   EntryPointGegen(GegenEnv*,C_Decl *);
   void emit() const;
+  virtual ~EntryPointGegen() {}
 };
 
 EntryPointGegen::EntryPointGegen(GegenEnv *e, C_Decl *f)
@@ -704,7 +719,7 @@ EntryPointGegen::EntryPointGegen(GegenEnv *e, C_Decl *f)
   use_cmixSR = false ;
   cmixSRisExp = false ;
   use_pendinglist = true ;
-  sharestrategy = "cmixFunNonshared" ;
+  DOcounter = 0 ;
 }
 
 void
@@ -753,9 +768,10 @@ EntryPointGegen::emit() const
   for ( Plist<C_Decl>::iterator rp2=paramit ; rp2 ; rp2++ )
     env.NameDynamicObject(*rp2,NULL);
 
-  env.ost << "  cmix_fun = cmixMkExp(cmixFunname);\n"
-    "  cmixPushFun(\"\", cmix_params,NULL,"
-    "cmix_fun,cmixNonshared,\n\t" ;
+  env.ost <<
+    "  cmixPushFun();\n"
+    "  cmix_fun = cmixMkExp(cmixFunname);\n"
+    "  cmixFunHeading(cmix_fun," ;
   residual_heading(env,fun,"","cmix_fun");
   env.ost << ");\n" ;
 
@@ -770,10 +786,10 @@ EntryPointGegen::emit() const
   s++ ;
   assert (!s) ;
 
-  // emit a return statement
+  // emit a return statement and end the function
   EmitJump(bb->exit());
-
-  env.ost << "  cmixPopFun(cmix_params,cmixNonshared);\n"
+  env.ost <<
+    "  cmixPopFun();\n"
     "}\n" ;
 
   if ( generator_name != cmixMain )
