@@ -38,12 +38,14 @@ static Plist<VarDecl>* enumconsts = new Plist<VarDecl>();
 static Plist<FunDef>* functions = new Plist<FunDef>();
 
 /////////// Namespace 3
-Scope<UserDecl> usernames;
+static Scope<UserDecl> usernames;
 // Names of structs/unions/enums.
 
-ListStack<UserDecl> userdecls;
-// Stack of user declarations; needed because UDs pop up in the middle of
-// ordinary declarations.
+static ListStack<UserDecl> userstack;
+// Stack of user declarations; needed to make user declarations in
+// parameter lists visible in the function body.
+
+static Plist<UserDecl> userdecls;
 
 /////////// Namespace 4
 static Scope<MemberDecl> membernames;
@@ -55,14 +57,28 @@ static ListStack<MemberDecl> members;
 
 //////// namespaces end ////////
 
+static void do_parameter_decay(Plist<VarDecl> *);
+static void type_mismatch(char const*, Type*, Position, Type*, Position);
+static void external_linker(ObjectDecl* ,bool is_defn);
 
-static int old_function = 0;
-// A value of means 1 when we are parsing the declaration list at the head of
-// an old function declaration. Set in function_defitinion and checked in the
-// function add_declaration.
+struct Declaration {
+  char const* const name ;
+  Type *const type ;
+  Position const pos ;
+  Parse_Storage storage ;
 
-static int reenter_scope_at_next_block = 0;
-// Set in rule function_definition and tested/reset in function enter_scope().
+  Declaration(Parse_GeneralType *gt, Parse_Declarator *dcl)
+    : name(dcl->id), type(dcl->maketype(gt)), pos(dcl->pos),
+      storage(gt->storage) {
+    assert(name != NULL );
+    delete dcl ;
+  }
+};
+
+static FunDef *current_function = NULL ;
+// set by begin_FunDef, used by end_FunDef
+
+static bool old_function = false ;
 
 static int last_decl_can_have_init = 0;
 // Controlled by add_declaration() and used by add_initializer().
@@ -96,8 +112,6 @@ unify_types(Type* t1, Type* t2)
   if (t1->realtype != t2->realtype)
     return false ;
   switch (t1->realtype) {
-  case EllipsisT:
-    return true ;
   case BaseT: {
     BaseType* b1 = (BaseType*) t1;
     BaseType* b2 = (BaseType*) t2;
@@ -125,12 +139,16 @@ unify_types(Type* t1, Type* t2)
     if ( f2->unspecified() ) {
       f2->params->clear() ;
       f2->params->splice(f2->params->begin(),*f1->params);
+      f2->varargs = f1->varargs ;
     } else if ( f1->unspecified() ) {
       f1->params->clear() ;
       f1->params->splice(f1->params->begin(),*f2->params);
+      f1->varargs = f2->varargs ;
     } else {
       if ( f1->params->size() != f2->params->size() )
         return false ;
+      if ( f1->varargs != f2->varargs )
+        return false ;
       double_foreach(p1, p2, *f1->params, *f2->params, Plist<Type>) {
         if ( !unify_types(*p1,*p2) )
           return false ;
@@ -246,6 +264,7 @@ make_or_reuse_struct(UserDecl *old,UserTag type, char const *name)
     }
   }
   StructDecl* d = new StructDecl(name,ccchere,type,NULL);
+  userstack.push_back(d);
   userdecls.push_back(d);
   usernames.insert(name, d);
   return d ;
@@ -319,6 +338,7 @@ cccParser::add_enum_decl(char const* name)
   }
   // Insert the enum.
   EnumDecl* d = new EnumDecl(name,ccchere,enumconsts);
+  userstack.push_back(d);
   userdecls.push_back(d);
   usernames.insert(name,d);
   // "Clear" the global list of enum constants.
@@ -332,103 +352,73 @@ cccParser :: find_enum_decl(char const* name)
   UserDecl* oldDecl = usernames.lookup(name);
   if (!oldDecl)
     Diagnostic(ERROR,ccchere) << "enum " << name << " undeclared" ;
+  else if(oldDecl->tag != Enum )
+    Diagnostic(ERROR,ccchere)
+      << name << " is a " << usertype2str(oldDecl->tag) << ", not an enum" ;
 }
 
 
 ///////////////////////////////////////////////
 
 
-BaseType*
-cccParser::make_basetype (Parse_Type* pt)
+Type*
+Parse_Type::gettype()
 {
-  BaseTypeTag tag = pt->determine_type();
+  BaseTypeTag tag = determine_type();
   BaseType *result = new BaseType(tag);
-  result->cv = pt->qualifiers;
+  result->cv = qualifiers;
   return result ;
 }
 
 Type*
-cccParser::make_typedeftype(Parse_TypedefType* tt)
+Parse_TypedefType::gettype()
 {
-  Type* t = tt->type->type;
-  // If there are no qualifiers, just return the unfolded type.
-  if ( tt->qualifiers.subsetof(constvol()) )
-    return t;
   // Put the qualifiers at the end of an array chain.
-  t = t->copy();
-  Type* handle = t;
+  Type* handle = type->type->copy();
+  Type* t = handle ;
   while (t->realtype == ArrayT) t = ((ArrayType*)t)->next;
   if ( t->realtype == FunT ) {
-    Diagnostic(WARNING,tt->type->pos)
+    Diagnostic(WARNING,type->pos)
       << "ignoring qualifiers on funcion type (through typedef)" ;
     return handle ;
   }
   // Check for multiple qualifiers.
-  if ( tt->qualifiers.cst && t->cv.cst )
-    Diagnostic(WARNING,tt->type->pos)
+  if ( qualifiers.cst && t->cv.cst )
+    Diagnostic(WARNING,type->pos)
       << "multiple const qualifiers when using this typedef";
-  if ( tt->qualifiers.vol && t->cv.vol )
-    Diagnostic(WARNING,tt->type->pos)
+  if ( qualifiers.vol && t->cv.vol )
+    Diagnostic(WARNING,type->pos)
       << "multiple volatile qualifiers when using this typedef";
-  t->cv += tt->qualifiers ;
+  t->cv += qualifiers ;
   return handle;
 }
 
-Parse_Decl*
-cccParser::make_decl (Parse_Type* pt)
+Type*
+Parse_UserType::gettype()
 {
-  return new Parse_Decl(make_basetype(pt),pt->storage);
-}
-
-Parse_Decl*
-cccParser::make_decl (Parse_UserType* ut)
-{
-  return new Parse_Decl(make_usertype(ut),ut->storage);
-}
-
-Parse_Decl*
-cccParser::make_decl (Parse_TypedefType* tt)
-{
-  return new Parse_Decl(make_typedeftype(tt),tt->storage);
-}
-
-
-UserType*
-cccParser::make_usertype (const Parse_UserType* const usert)
-{
-  UserDecl* def = usernames.lookup(usert->id);
+  UserDecl* def = usernames.lookup(id);
   UserType* result ;
   assert(def != NULL);
-  switch (usert->type) {
+  switch (type) {
   case Struct: result = new StructType(def); break;
   case Union:  result = new UnionType(def); break;
   case Enum:   result = new EnumType(def); break;
   default: 
-    Diagnostic(INTERNAL,RIGHT_HERE) << "cccParser::make_usertype";
+    Diagnostic(INTERNAL,RIGHT_HERE) << "Parse_UserType::gettype";
     return NULL;
   }
-  result->cv = usert->qualifiers ;
+  result->cv = qualifiers ;
   return result ;
 }
 
 void
 cccParser::enter_scope (void)
 {
-  if (reenter_scope_at_next_block) {
-    if (debug^5) debug << "[reenter scope]";
-    reenter_scope_at_next_block = 0;
-    names.reenter_previous_scope();
-    objects.reenter();
-    usernames.reenter_previous_scope();
-    userdecls.reenter();
-  }
-  else {
-    if (debug^5) debug << "[enter scope]";
-    names.enter_scope();
-    objects.enter();
-    usernames.enter_scope();
-    userdecls.enter();
-  }
+  if (debug^5) debug << "[enter scope]";
+  names.enter_scope();
+  objects.enter();
+  usernames.enter_scope();
+  userstack.enter();
 }
 
 void
@@ -440,17 +430,108 @@ cccParser::leave_scope (void)
   if (debug^5) debug << "[leave scope]";
   
   usernames.leave_scope();
-  
-  // Ugly hack: in case there are userdecls left here, we export
-  // them to the enclosing scope, so they won't be forgot in the
-  // c2core phase.
-  Plist<UserDecl> *olduserdecls = userdecls.get_list() ;
-  userdecls.leave();
-  foreach(i,*olduserdecls,Plist<UserDecl>)
-      userdecls.push_back(*i);
-  delete olduserdecls ;
+  userstack.leave();
 }
 
+Plist<Parse_Typemod>*
+cccParser::leave_parameter_scope(bool varargs)
+{
+  Plist<VarDecl> *params = objects.get_list() ;
+  Plist<UserDecl> *userdefs = userstack.get_list() ;
+  leave_scope() ;
+  
+  // A single anonymous void parameter is removed, unless there are varargs
+  if( !varargs && params->size() == 1 && params->front()->type->isVoid ) {
+    delete params->front() ;
+    params->pop_front() ;
+  }
+  do_parameter_decay(params);
+  
+  // Warn about usertypes declared here
+  foreach(i,*userdefs,Plist<UserDecl>) {
+    Diagnostic(WARNING,i->pos) << usertype2str(i->tag) << " " << i->name
+                               << " declared in parameter list" ;
+  }
+
+  return new Plist<Parse_Typemod>
+    (new Parse_Funtype(params,varargs,userdefs,ccchere));
+}
+
+void
+cccParser::enter_old_parameter_scope()
+{
+  enter_scope() ;
+  old_function = true ;
+}
+
+void
+cccParser::leave_old_parameter_scope(Parse_Typemod *tm)
+{
+  Parse_Funtype *fun = tm->me_as_function() ;
+  assert(fun);
+
+  // move the usertype declarations (there are none in
+  // an old-style parameter list)
+  assert(fun->usrdecls->empty());
+  delete fun->usrdecls ;
+  fun->usrdecls = userstack.get_list();
+  
+  // produce the final parameter declaration list
+  Plist<VarDecl> *result = new Plist<VarDecl> ;
+  Plist<VarDecl> *typed = objects.get_list() ;
+  do_parameter_decay(typed);
+  foreach(iname,*fun->params,Plist<VarDecl>) {
+    for ( Plist<VarDecl>::mutable_iterator ityped = typed->begin() ;
+          true ; ityped++ ) {
+      if( !ityped ) {
+        result->push_back(*iname);
+        break ;
+      }
+      if( strcmp(ityped->name,iname->name) == 0 ) {
+        result->push_back(*ityped);
+        delete *iname ;
+        typed->erase(ityped);
+        break ;
+      }
+    }
+  }
+
+  // complain about parameters that have not yet been moved to
+  // the final parameter list
+  // Add them to the parameters anyway, so the lack of them won't
+  // cause trouble during the rest of the parsing phase.
+  foreach(i,*typed,Plist<VarDecl>) {
+    Diagnostic(ERROR,i->pos) << i->name
+                             << " is not mentioned in the parameter list" ;
+    result->push_back(*i);
+  }
+
+  // clean up and return
+  delete typed ;
+  delete fun->params ;
+  fun->params = result ;
+
+  old_function = false ;
+  leave_scope();
+}
+
+static void
+do_parameter_decay(Plist<VarDecl> *parlist)
+{
+  foreach(i,*parlist,Plist<VarDecl>) {
+    Type *t = i->type ;
+    switch(t->realtype) {
+    case ArrayT:
+      i->type = new PtrType(((ArrayType*)t)->next) ;
+      break ;
+    case FunT:
+      i->type = new PtrType(t);
+      break ;
+    default:
+      break ;
+    }
+  }
+}
 
 void
 cccParser::members_enter_scope (void)
@@ -475,17 +556,20 @@ cccParser::make_typedef (const char* name)
   return new Parse_TypedefType(tdef);
 }
 
-MemberDecl*
-cccParser::add_member (Parse_Decl *d, Expr* bits)
+void
+cccParser::add_member(Parse_GeneralType *gt, Parse_MemberId *mid)
 {
+  if( mid->p_id == NULL )
+    mid->p_id = new Parse_Declarator(fresh_anonymous_name(),ccchere);
+  const Declaration d(gt,mid->p_id);
   // Check that the name was not seen before.
-  if (membernames.lookup_local(d->name)) {
-    Diagnostic(ERROR,ccchere) << "redeclaration of member " << d->name ;
+  if (membernames.lookup_local(d.name)) {
+    Diagnostic(ERROR,ccchere) << "redeclaration of member " << d.name ;
   }
-  VarDecl* var = new VarDecl(d->name,d->pos,d->type,NoLinkage,VarMu);
-  MemberDecl* decl = new MemberDecl(var,bits);
+  VarDecl* var = new VarDecl(d.name,d.pos,d.type,NoLinkage,VarMu);
+  MemberDecl* decl = new MemberDecl(var,mid->bits);
   // Make the name known.
-  membernames.insert(d->name, decl) ;
+  membernames.insert(d.name, decl) ;
   // Put the actual definition in the back of the list of members. This list
   // is collected in update_struct_decl() and anonymous_struct_decl(). 
   members.push_back(decl); 
@@ -494,7 +578,7 @@ cccParser::add_member (Parse_Decl *d, Expr* bits)
     decl->show(debug);
     debug << "\n";
   }
-  return decl;
+  delete mid;
 }
 
 // Construct an enum constant and put it in current scope.
@@ -542,141 +626,137 @@ cccParser::add_enumconst (char* n, Position pos, Expr *value)
   return decl;
 }
 
-FunDef*
-cccParser::make_final_FunDef (Parse_Decl* board,
-                              Plist<VarDecl>* ds,
-                              Plist<UserDecl>* us)
+static bool
+findPrototype(Linkage link, char const* id, Type* type, Position pos)
 {
-  if ( !board->type->isFun ) {
-    Diagnostic(ERROR,ccchere) << board->name
-                              << " is not a function so it cannot "
-                              << " have a body" ;
-    return NULL ;
+  ObjectDecl *proto = names.lookup(id) ;
+  if( proto == NULL )
+    return false ;
+
+  if ( link == Internal && proto->linkage == External ) {
+    Diagnostic d(ERROR,pos);
+    d << id << " redeclared as static" ;
+    d.addline(proto->pos) << "(previous declaration here)" ;
   }
+
+  // Check that the declarations match.
+  if (!unify_types(type,proto->type)) {
+    type_mismatch(id,proto->type,proto->pos,type,pos);
+    return false ;
+  }
+  assert(proto->tag==Fundef);
+  current_function = (FunDef*) proto;
+  if (current_function->stmts != NULL) {
+    Diagnostic d(ERROR,pos);
+    d << "redefinition of function " << id;
+    d.addline(current_function->pos) << "(previous definition here)" ;
+    return false ;
+  }
+  
+  // Use the existing object but transfer the type and position
+  current_function->type = type;
+  current_function->pos = pos;
+  assert(current_function->decls == NULL);
+
+  // if this prototype is not the lead element in the union, promote it
+  if( current_function->refer_to != NULL ) {
+    current_function->refer_to = NULL ;
+    external_linker(current_function,true);
+    assert(current_function->refer_to == NULL);
+  }
+  return true ;
+}
+
+void
+cccParser::begin_FunDef(Parse_GeneralType *gt, Parse_Declarator *name)
+{
+  assert(current_function == NULL);
+
+  // Make sure the type of the declaration is a function, and
+  // extract the function modifier
+  Parse_Funtype *funmod = NULL ;
+  if( !name->typemods.empty() )
+    funmod = name->typemods.back()->me_as_function() ;
+  if( funmod == NULL ) {
+    Diagnostic err(ERROR,name->pos);
+    err << name->id
+        << " is not a function, so it can't have a body" ;
+    funmod = new Parse_Funtype(new Plist<VarDecl>(),false,
+                               new Plist<UserDecl>,name->pos);
+    name->typemods.push_back(funmod);
+  }
+  funmod->pos = name->pos ;
+
+  // The function head "foo()" has been parsed as "foo(...)" because
+  // that is what that kind of prototypes mean. It is different for
+  // function *definitions* -- adjust.
+  if ( funmod->params->empty() )
+    funmod->varargs = false ;
+
+  // Create the function's type
+  Type* t = name->maketype(gt);
+
+  // Select the linkage
   Linkage link = External;
-  switch (board->storage) {
+  switch (gt->storage) {
   case P_NoStorageClass:
   case P_Extern:          link=External; break;
   case P_Static:          link=Internal; break;
   case P_Typedef:         
   case P_Register:
   case P_Auto:
-    Diagnostic(ERROR,board->pos) << board->name <<
+    Diagnostic(ERROR,name->pos) << name->id <<
       "() has a bad storage class specifier" ;
   }
-  return new FunDef(board->name,board->pos,board->type,link,ds,us);
-}
 
-FunDef*
-cccParser :: common_add_Fun (FunDef* f)
-{
-  // check for prototypes
-  ObjectDecl *proto = names.lookup(f->name) ;
-  if ( proto != NULL ) {
-    // reuse a prototype.
-    if ( f->linkage == External )
-      f->linkage = proto->linkage ;
-    if ( f->linkage != proto->linkage ) {
-      Diagnostic d(ERROR,f->pos);
-      d << f->name << " redeclared with different linkage" ;
-      d.addline(proto->pos) << "(previous declaration here)" ;
-    }
-    // Check that the declarations match.
-    if (!unify_types(f->type,proto->type))
-      type_mismatch(f->name,proto->type,proto->pos,f->type,f->pos);
-    assert(proto->tag==Fundef);
-    FunDef* x = (FunDef*) proto;
-    if (x->stmts != NULL) {
-      Diagnostic d(ERROR,f->pos);
-      d << "redefinition of function " << x->name;
-      d.addline(x->pos) << "(previous definition here)" ;
-    }
-    // Use the existing object but transfer body and position.
-    if (debug^5) {
-      debug << "add_fundef: using existing function "
-            << x->name << "\n";
-    }
-    x->pos = f->pos;
-    x->decls = f->decls;
-    x->userdecls = f->userdecls;
-    x->stmts = f->stmts;
-    x->effects = f->effects ;
-    x->calltime = f->calltime ;
-    delete f ;
-    // if this prototype is not the lead element in the union, promote it
-    if( x->refer_to != NULL ) {
-      x->refer_to = NULL ;
-      external_linker(x,true);
-    }
-    reenter_scope_at_next_block = 1;
-    return x;
+  // Create a FunDef for the new function, or reuse a prototype
+  if( findPrototype(link,name->id,t,name->pos) ) {
+    current_function->decls = funmod->params ;
   } else {
-    // There is no prototype in scope
-    // Make the name and definion known.    
-    names.insert(f->name,f);
-    functions->push_back(f);
-    if ( f->linkage == External )
-      external_linker(f,true);
-    // Signal that forthcomming declarations in the body of the
-    // function belongs to the same name scope.
-    reenter_scope_at_next_block = 1;
-    if (debug^4) {
-      debug << "add_FunDef: ";
-      f->show(debug);
-      debug << "\n";
-    }
-    return f;
+    current_function
+      = new FunDef(name->id,name->pos,t,link,funmod->params);
+    names.insert(name->id,current_function);
+    functions->push_back(current_function);
+    if( current_function->linkage == External )
+      external_linker(current_function,true);
   }
-}
 
-FunDef*
-cccParser :: add_FunDef (Parse_Decl* board)
-{
-  // Create a function definition with all the current parameters
-  // from the object scope that has just been left.
-  reenter_scope_at_next_block = 1;
+  // Open a scope for the function body, and insert the parameters into
+  // it (checking that they all have names and none of them have
+  // initializers)
   enter_scope();
-  FunDef* f = make_final_FunDef(board,objects.get_list(),userdecls.get_list());
-  leave_scope();
-  return common_add_Fun(f);
+  bool metAnonymous = false ;
+  foreach(i,*funmod->params,Plist<VarDecl>) {
+    if( i->init )
+      Diagnostic(ERROR,i->pos) << "Parameter " << i->name
+                               << " cannot have initializer" ;
+    if( i->name[0] == '$' )
+      metAnonymous = true ;
+    else
+      names.insert(i->name,*i);
+  }
+  if( metAnonymous )
+    Diagnostic(WARNING,name->pos)
+      << "Standard C does not allow anonymous parameters" ;
+  foreach(i,*funmod->usrdecls,Plist<UserDecl>)
+    usernames.insert(i->name,*i);
+
+  // protect the parameters from being deallocated
+  name->typemods.pop_back();
+  funmod->params = NULL ;
+  delete funmod ;
+  delete name ;
+  delete gt ;
 }
 
 void
-cccParser::old_style_decl_list (void)
+cccParser::end_FunDef(Stmt* stmt)
 {
-  old_function=1;
-  reenter_scope_at_next_block = 1;
-  enter_scope();
-}
-
-FunDef*
-cccParser::old_style_add_FunDef (Parse_Decl* board)
-{
-  old_function=0;
-  // Create a function definition with all the current parameters
-  // from the object scope that has just been left.
-  reenter_scope_at_next_block = 1;
-  enter_scope();
-  // Collect the types for the arguments. 
-  Plist<VarDecl>* parameters = objects.get_list();
-  Plist<Type>* types = new Plist<Type>();
-  Plist<VarDecl>::iterator first = parameters->begin();
-  while(first) {
-    // If no type is specified it gets to be integer.
-    if ((*first)->type == NULL)
-      (*first)->type = BaseType::get(Int);
-    types->push_back((*first)->type);
-    first++;
-  }
-  board->type = new FunType(types,board->type);
-  FunDef* f = make_final_FunDef(board,parameters,userdecls.get_list());
+  assert(current_function != NULL);
+  current_function->stmts = stmt ;
   leave_scope();
-  //     if (debug^4) {
-  //         debug << "old_fun: ";
-  //         f->show(debug);
-  //         debug << "\n";
-  //     }
-  return common_add_Fun(f);
+  goto_labels.clear();
+  current_function = NULL ;
 }
 
 void
@@ -718,58 +798,22 @@ cccParser::add_goto (char* label, Position p)
   }
 }
 
-void cccParser :: update_arg_old_fun_decl(Parse_Decl* d)
-{
-  // Find the dummy declaration for the parameter.
-  ObjectDecl* oldDecl = names.lookup_local(d->name);
-  if (!oldDecl) {
-    Diagnostic di(ERROR,ccchere);
-    di << "old style function declaration:" ;
-    di.addline() << "no such parameter " << d->name ;
-  }
-  else {
-    assert(oldDecl->tag == Vardecl);
-    VarDecl* old = (VarDecl*) oldDecl;
-    if (old->type) {
-      Diagnostic di(ERROR,ccchere);
-      di << "old style function declaration:" ;
-      di.addline() << "redeclaration of " << d->name ;
-      if (debug^2) {
-        di.addline() << "old type was ";
-        old->type->show(di);
-        di << "]";
-      }
-      return ;
-    }
-    // Update the placeholder.
-    old->pos = d->pos;
-    old->type = d->type;
-    if (debug^4) {
-      debug << "update parameter: ";
-      old->show(debug);
-      debug << "\n";
-    }
-    // The varmode is already set: old-style parameters start out
-    // their life as ordinary 'noclass' declarations with type NULL!
-  }
-}
-
-void
-cccParser::add_noclass (Parse_Decl* d)
+static void
+add_noclass (Declaration const& d)
 {
   // this is only used for *local* noclass definitions of *objects*.
-  VarDecl* decl = new VarDecl(d->name, d->pos,d->type,NoLinkage,VarIntAuto);
+  VarDecl* decl = new VarDecl(d.name, d.pos,d.type,NoLinkage,VarIntAuto);
   if (debug^4) {
     debug << "add_noclass: ";
     decl->show(debug);
     debug << "\n";
   }
-  names.insert(d->name, decl);
+  names.insert(d.name, decl);
   objects.push_back(decl);
 }
 
-void
-cccParser::external_linker(ObjectDecl *d,bool is_defn)
+static void
+external_linker(ObjectDecl *d,bool is_defn)
 {
   static SymbolTable<ObjectDecl> externals;
 
@@ -799,16 +843,16 @@ cccParser::external_linker(ObjectDecl *d,bool is_defn)
   d->refer_to = prev ;
 }
 
-void
-cccParser::add_extern(Parse_Decl* d, bool tentative_defn)
+static void
+add_extern(Declaration const& d, bool tentative_defn)
 {
-  ObjectDecl* ex = names.lookup_global(d->name);
+  ObjectDecl* ex = names.lookup_global(d.name);
   if (ex != NULL) {
     // If there is a visible file scope definition we use it and
     // inherit its linkage.
     // Check that the declarations match.
-    if ( !unify_types(ex->type,d->type) ) {
-      type_mismatch(d->name,ex->type,ex->pos,d->type,d->pos);
+    if ( !unify_types(ex->type,d.type) ) {
+      type_mismatch(d.name,ex->type,ex->pos,d.type,d.pos);
       return ;
     }
     // if this is a tentative definition (which only occurs
@@ -824,24 +868,24 @@ cccParser::add_extern(Parse_Decl* d, bool tentative_defn)
       }
     }
     if (debug^5) {
-      debug << "add_extern: using existing object for " << d->name << endl ;
+      debug << "add_extern: using existing object for " << d.name << endl ;
     }
     // Insert a placeholder (pointing to the global).
-    names.insert(d->name,ex);
+    names.insert(d.name,ex);
   } else {
     // There is no visible global definition for this. Make a new
     // one; with external linkage.
     ObjectDecl* decl;
     // Function prototypes should eventually be a real function,
     // so they are inserted as a function.
-    if (d->type->isFun) {
-      FunDef* fun = new FunDef(d->name, d->pos,d->type,External,NULL,NULL);
+    if (d.type->isFun) {
+      FunDef* fun = new FunDef(d.name, d.pos,d.type,External,NULL);
       functions->push_back(fun);
       decl = fun;
     } else {
       VariableMode thevarmode = tentative_defn ? VarIntAuto : VarExtDefault ;
       VarDecl* var
-        = new VarDecl(d->name, d->pos,d->type,External,thevarmode);
+        = new VarDecl(d.name, d.pos,d.type,External,thevarmode);
       objects.push_back_global(var);
       decl = var;
     }
@@ -850,21 +894,21 @@ cccParser::add_extern(Parse_Decl* d, bool tentative_defn)
       decl->show(debug);
       debug << "\n";
     }
-    names.insert(d->name, decl);
+    names.insert(d.name, decl);
     external_linker(decl,tentative_defn);
   }
 }
 
-void
-cccParser::add_static(Parse_Decl* d)
+static void
+add_static(Declaration const& d)
 {
   ObjectDecl* decl;
-  if (d->type->isFun) {
-    FunDef* fun = new FunDef(d->name,d->pos,d->type,Internal,NULL,NULL);
+  if (d.type->isFun) {
+    FunDef* fun = new FunDef(d.name,d.pos,d.type,Internal,NULL);
     functions->push_back(fun);
     decl = fun;
   } else {
-    VarDecl* var = new VarDecl(d->name,d->pos,d->type,Internal,VarIntAuto);
+    VarDecl* var = new VarDecl(d.name,d.pos,d.type,Internal,VarIntAuto);
     var->static_local = true ;
     objects.push_back(var);
     decl = var;
@@ -874,45 +918,51 @@ cccParser::add_static(Parse_Decl* d)
     decl->show(debug);
     debug << "\n";
   }
-  names.insert(d->name, decl);
+  names.insert(d.name, decl);
 }
 
 // Construct a final declaration, check the integrity of it,
 // and put it in current scope.
-void cccParser::add_declaration (Parse_Decl *d)
+void
+cccParser::add_declaration (Parse_GeneralType *gt, Parse_Declarator *dcl)
 {
-  assert(d->name != NULL);
+  // the name can be NULL if the "declaration" is a prototype parameter
+  if( dcl->id == NULL )
+    dcl->id = fresh_anonymous_name();
 
+  Declaration d(gt,dcl);
+  
   last_decl_can_have_init = 1;
-  if (old_function) {
-    last_decl_can_have_init = 0;
-    update_arg_old_fun_decl(d);
-    return ;
+  if( old_function ) {
+    last_decl_can_have_init = false ;
+    if( d.storage != P_Register && d.storage != P_NoStorageClass )
+      Diagnostic(ERROR,ccchere)
+        << "storage classes are not allowed for parameters" ;
   }
 
   // Find a previous declaration if such exists.
-  ObjectDecl* oldDecl = names.lookup_local(d->name);
+  ObjectDecl* oldDecl = names.lookup_local(d.name);
   // If we're at anything but file scope, it is not allowed to
   // redeclare a name.
   if ( names.get_level() > 0 && oldDecl ) {
     Diagnostic di(ERROR,ccchere) ;
-    di << "redeclaration of " << d->name ;
+    di << "redeclaration of " << d.name ;
     di.addline(oldDecl->pos) << "(previous declaration here)"  ;
     return ;
   }
 
-  if (d->type != NULL && d->type->isFun) {
+  if (d.type != NULL && d.type->isFun) {
     last_decl_can_have_init = 0;
     // Function prototypes in non-global scope can only have extern or no
     // storage specifier.
-    if ( names.get_level() > 0 && d->storage != P_Extern ) {
+    if ( names.get_level() > 0 && d.storage != P_Extern ) {
       Diagnostic(ERROR,ccchere) << "prototype can only have external linkage";
-      d->storage = P_Extern ;
+      d.storage = P_Extern ;
     }
   }
   // If the declaration is extern or static we put it in the
   // global list of declarations. 
-  switch (d->storage) {
+  switch (d.storage) {
   case P_Extern:
     add_extern(d,false);
     break;
@@ -922,15 +972,15 @@ void cccParser::add_declaration (Parse_Decl *d)
   case P_Typedef:
     last_decl_can_have_init = 0;
     if (oldDecl)
-      Diagnostic(ERROR,ccchere) << "redeclaration of " << d->name;
+      Diagnostic(ERROR,ccchere) << "redeclaration of " << d.name;
     else {
-      VarDecl *decl = new VarDecl(d->name, d->pos,d->type,Typedef,VarMu);
+      VarDecl *decl = new VarDecl(d.name, d.pos,d.type,Typedef,VarMu);
       if (debug^4) {
         debug << "add_typedef: ";
         decl->show(debug);
         debug << "\n";
       }
-      names.insert(d->name, decl);
+      names.insert(d.name, decl);
       objects.push_back(decl);
     }
     break;
@@ -938,7 +988,7 @@ void cccParser::add_declaration (Parse_Decl *d)
   case P_Auto:
     if (names.get_level() == 0)
       Diagnostic(ERROR,ccchere) << "register or auto specified for "
-                                   "top-level declaration " << d->name ;
+                                   "top-level declaration " << d.name ;
     add_noclass(d);
     break;
   case P_NoStorageClass:
@@ -948,13 +998,13 @@ void cccParser::add_declaration (Parse_Decl *d)
     // the Standard, 6.7.2
     if ( names.get_level()==0 ) {
       // NULL types should occur only in old-style parameter lists.
-      assert(d->type != NULL);
-      add_extern(d,!d->type->isFun);
+      assert(d.type != NULL);
+      add_extern(d,!d.type->isFun);
     } else
       add_noclass(d);
     break;
   default:
-    Diagnostic(INTERNAL,d->pos) << "unknown storage class";
+    Diagnostic(INTERNAL,d.pos) << "unknown storage class";
   }
 }
 
@@ -982,10 +1032,10 @@ cccParser :: add_initializer (Init* init)
   }
 }
 
-void
-cccParser::type_mismatch(char const *n,
-                         Type *t1,Position p1,
-                         Type *t2, Position p2)
+static void
+type_mismatch(char const *n,
+              Type *t1,Position p1,
+              Type *t2, Position p2)
 {
   Diagnostic d(ERROR,Position());
   d << "conflicting types given for " << n << ':' ;
@@ -1019,6 +1069,6 @@ cccParser :: get_program()
   CProgram* pgm = new CProgram;
   pgm->functions = functions;
   pgm->definitions = objects.get_list();
-  pgm->usertypes = userdecls.get_list();
+  pgm->usertypes = &userdecls ;
   return pgm;
 }
