@@ -159,14 +159,6 @@ Constvol(Plist<Output>* os, Type* t)
     }
 }
 
-
-Output*
-EllipsisType::output(OutCpgm&, Output* middle, bool /*precedence*/)
-{
-    assert(middle==NULL && cv.cst==0 && cv.vol==0);
-    return new Output("...",OKeyword);
-}
-
 Output* AbsType::output(OutCpgm &env, Output* middle, bool) {
     Plist<Output>* os = new Plist<Output>(new Output(name,OTypename));
     Constvol(os,this);
@@ -244,6 +236,8 @@ FunType::output(OutCpgm &env, Output* middle, bool precedence)
     parlist.sep.push_back(env.decorate(btannos,comma));
     foreach(i,*params,Plist<Type>)
         parlist += i->output(env);
+    if ( varargs )
+      parlist += ellipsis ;
     os->push_back(parlist.inter());
     os->push_back(rparen);
     return ret->output(env, new Output(Output::Consistent,os), false);
@@ -398,12 +392,6 @@ FunDef::output(OutCpgm &env)
       env.mixreason = saved ;
     } else
         os->push_back(semi);
-    // Remember: these declarations originate from the parameters.
-    if ( userdecls != NULL )
-        foreach(ud,*userdecls,Plist<UserDecl>) {
-            os->push_front(newline);
-            os->push_front(ud->output(env));
-        } 
     if (debug^1) debug << "]";
     return new Output(Output::Consistent,os);
 }
@@ -807,8 +795,6 @@ CompoundStmt::output(OutCpgm &env)
     // Make an indented block out of the statements.
     OutputList body(0,Output::Consistent);
     body.sep.push_back(newline);
-    foreach(ut,*usertypes,Plist<UserDecl>)
-        body += ut->output(env);
     foreach(o,*objects,Plist<VarDecl>)
         body += oconcat(o->output(env),semi);
     foreach(s,*stmts,Plist<Stmt>)
