@@ -9,7 +9,6 @@
  *      terms; see the file COPYING.cmix for details.
  */
 
-#include <cmixconf.h>
 #include <stdlib.h>
 #include <string.h>
 #include <signal.h>
@@ -21,6 +20,7 @@ int cmixSpeclibAbortSignal = SIGKILL ;
 
 void cmixFatal(char *err)
 {
+    fflush(NULL);
     fprintf(stderr, "C-Mix fatal error: %s\n", err);
     cmixPrintMemusage();
     raise(cmixSpeclibAbortSignal);
