@@ -4,7 +4,7 @@
  *                Henning Makholm
  *      Content:  C-Mix speclib: memoization
  *
- *      Copyright  1999. The TOPPS group at DIKU, U of Copenhagen.
+ *      Copyright  1999-2000. The TOPPS group at DIKU, U of Copenhagen.
  *      Redistribution and modification are allowed under certain
  *      terms; see the file COPYING.cmix for details.
  */
@@ -13,12 +13,17 @@
 #include <sys/types.h>
 #include <memory.h>
 #include <string.h>	/* for memcmp */
-#include <assert.h>
 #define cmixSPECLIB_SOURCE
 #include "speclib.h"
+#include "code.h"  /* for struct cmixFun */
 
-#define VOIDPTR_PLUS(a,b) ((void*)((char*)(a)+(b)))
-#define VOIDPTR_MINUS(a,b) ((char*)(a)-(char*)(b))
+struct cmixDataObjectCopy {
+  void *copy;				/* Copied data                */
+  size_t size;		        	/* Size of the copied object  */
+  struct cmixDataObject *obj;		/* The original object desc.  */
+  long offset ;		        	/* offset always memoized     */
+  struct cmixDataObjectCopy *next;	/* Next element in the list   */
+};
 
 typedef struct HeapObject {
   cmixDataObject obj ;
@@ -110,7 +115,8 @@ void cmixPopLocals(int n)
 
 static cmixMemoWhat global_what ;
 
-static void clear_visit_flags()
+static void
+clear_visit_flags()
 {
   int i ;
   HeapObject *hop ;
@@ -125,24 +131,25 @@ static void clear_visit_flags()
     hop->obj.visit = 0;
 }
 
+static int
+setinwhat(cmixDataObject *d)
+{
+  unsigned id = d->ID ;
+  return ( (unsigned char)(global_what[id/8]) & (1 << (id&7)) ) != 0 ;
+}
+
 static int
 maycopy(cmixDataObject *d)
 {
   if ( d->visit ) {
-    DEBUG_MEM(("already visited - ID %ld\n",ID));
+    DEBUG_MEM(("already visited - ID %ld\n",d->ID));
     return 0 ;
   }
   d->visit = 1 ;
-  if ( global_what )
-    switch( global_what[d->ID] ) {
-    case '0':
-      DEBUG_MEM(("not in use - ID %ld\n",d->ID));
-      return 0 ;
-    case '1':
-      break ;
-    default:
-      cmixFatal("strange memoisation data");
-    }
+  if ( !setinwhat(d) ) {
+    DEBUG_MEM(("not in use - ID %ld\n",d->ID));
+    return 0 ;
+  }
   return 1 ;
 }
 
@@ -276,34 +283,16 @@ copy_objects(cmixDataObject objs[])
   return copies;
 }
 
-cmixSavedState
-cmixCopyState(cmixDataObject locals[], cmixMemoWhat what)
+static cmixSavedState
+cmixCopyState(cmixDataObject locals[])
 {
   cmixDataObjectCopy *temp ;
-  global_what = what ;
 
   clear_visit_flags();
   temp = copy_objects(globals);
   return append(temp,copy_objects(locals));
 }
 
-cmixSavedState
-cmixCopyLightState(cmixDataObject locals[], cmixMemoWhat what)
-{
-  cmixDataObjectCopy *copies = NULL;
-  unsigned i ;
-
-  global_what = what ;
-  clear_visit_flags();
-  for(i = 0; locals[i].obj != NULL; i++) {
-    DEBUG_MEM(("copy_light_object: state->locals[%d] = %p\n",
-               i, (void*)&locals[i]));
-    copies = append(copies,do_copy_object(&locals[i],NULL));
-  }
-  
-  return copies;
-}
-
 /* Make a state copy containing the new values of those objects
  * memoised as the initial state for the function that have changed.
  * Local objects in the function are not copied; it is irrelevant
@@ -317,7 +306,7 @@ cmixCopyLightState(cmixDataObject locals[], cmixMemoWhat what)
  *   an object and then forget the pointer to the changed object, though
  *   it may still be visible to the caller.
  */
-cmixSavedState
+static cmixSavedState
 cmixExtractEndState(cmixDataObjectCopy *beginstate, cmixDataObject locals[])
 {
   cmixDataObjectCopy *restore = NULL ;
@@ -337,6 +326,7 @@ cmixExtractEndState(cmixDataObjectCopy *beginstate, cmixDataObject locals[])
   do {
     /* local objects are irrelevant; optimize away unchanged objects */
     if ( !copy->obj->visit &&
+         setinwhat(copy->obj) &&
          memcmp(copy->obj->obj, copy->copy, copy->size) != 0 ) {
       /* OK, make a copy of this object */
       cmixDataObjectCopy *newcopy = NEW(cmixDataObjectCopy) ;
@@ -354,7 +344,7 @@ cmixExtractEndState(cmixDataObjectCopy *beginstate, cmixDataObject locals[])
  * Restore a saved state						    *
  *==========================================================================*/
 
-void
+static void
 cmixRestoreState(cmixSavedState copy)
 {
   cmixDataObjectCopy *c = copy ;
@@ -370,11 +360,13 @@ cmixRestoreState(cmixSavedState copy)
     } while (c != copy);
 }
 
+#if 0
+
 /*==========================================================================*
  * Delete a saved state							    *
  *==========================================================================*/
 
-void
+static void
 cmixFreeState(cmixSavedState c)
 {
   cmixDataObjectCopy *ob, *next;
@@ -390,6 +382,8 @@ cmixFreeState(cmixSavedState c)
   }
 }
 
+#endif
+
 /*==========================================================================*
  * Compare saved state to current state					    *
  *==========================================================================*/
@@ -484,15 +478,13 @@ compare_one(cmixDataObject *d,unsigned length,
   return d->follow_pointers(compare_referenced_object,copy,d->obj,length);
 }
 
-int
-cmixCmpState(cmixDataObject locals[], cmixSavedState copy, cmixMemoWhat what)
+static int
+cmixCmpState(cmixDataObject locals[], cmixSavedState copy)
 {
   cmixDataObjectCopy *cpy ;
   unsigned i ;
   
-  global_what = what ;
-  DEBUG_MEM(("cmixCmpState: state: %p, copy: %p\n",
-             (void*)state, (void*)copy));
+  DEBUG_MEM(("cmixCmpState: copy: %p\n",(void*)copy));
   /* The empty state always matches */
   if (copy==NULL)
     return 1;
@@ -525,20 +517,67 @@ cmixCmpState(cmixDataObject locals[], cmixSavedState copy, cmixMemoWhat what)
   return 0;
 }
 
-int
-cmixCmpLightState(cmixSavedState copy)
-{
-  cmixDataObjectCopy *cpy = copy ;
+/*==========================================================================*
+ * Interface for function memoization					    *
+ *==========================================================================*/
 
-  DEBUG_MEM(("cmixCmpLightState: copy: %p\n", (void*)copy));
-  /* The empty state always matches */
-  if (copy!=NULL)
-    do {
-      if ( 0 != memcmp(cpy->copy,cpy->obj->obj,cpy->size) )
-        return 0 ;
-      cpy = cpy->next ;
-    } while ( cpy != copy );
-  return 1 ;
+static cmixSavedState last_nonfound = NULL ;
+struct cmixFunPool *last_pool = NULL ;
+
+/* Check whether a function has been specialized to this state.
+ * If found then return a resiudal expression containing the name of the
+ * residual function. Otherwise prepare for creating the new function
+ */
+Code
+cmixFindOrPushFun(struct cmixFunPool *pool,
+                  cmixDataObject params[],cmixMemoWhat what)
+{
+  struct cmixFun *f ;
+
+  global_what = what ;
+  assert(last_pool == NULL);
+    
+  /* Go through residual functions and check... */
+  for ( f = pool->list ; f ; f = f->next_share ) {
+    if ( !cmixCmpState(params, f->state) )
+      continue ;
+    if ( f->end_state )
+      cmixRestoreState(f->end_state);
+    f->shared = 1 ;
+    return f->resname;
+  }
+
+  last_nonfound = cmixCopyState(params);
+  last_pool = pool ;
+  return NULL ;
+}
+
+void
+cmixPushFun(void)
+{
+  assert(last_pool == NULL);
+  /* last_pool = NULL */
+  last_nonfound = NULL ;
+}
+
+void
+cmixFinishPushedFun(struct cmixFun *f)
+{
+  if( last_pool ) {
+    f->next_share = last_pool->list ;
+    last_pool->list = f ;
+  }
+  f->state = last_nonfound ;
+  f->end_state = 0 ;
+  f->shared = 0 ;
+  last_pool = NULL ;
+}
+
+void
+cmixSaveExitState(cmixDataObject locals[],cmixMemoWhat what)
+{
+  global_what = what ;
+  cmixCurFun->end_state = cmixExtractEndState(cmixCurFun->state,locals);
 }
 
 /*==========================================================================*
