@@ -12,10 +12,10 @@
 #include <stdlib.h>
 #include <string.h>
 #include <ctype.h>
+#include <stdarg.h>
 #define cmixSPECLIB_SOURCE
 #include "speclib.h"
 #include "code.h"
-#include <assert.h>
 
 /***************************************************************************/
 /* Adding taboos to a name table                                           */
@@ -278,6 +278,7 @@ printInner(struct cmixExprInner *e,struct cmixNameRec names[],FILE *f)
   unsigned yeslevel = 0 ;
   int inPling = 0 ;
   int printing = 1 ;
+  if ( f == NULL ) return;
   for ( pc = e->string ; *pc ; pc++ )
     switch(*pc) {
     case '#':
@@ -379,9 +380,13 @@ printExpr(union cmixExpr *e,struct cmixNameRec names[],FILE *f)
   case LiftLongDouble:
     e->lift_float.write(f,e->common.tag,e->lift_float.data);
     break ;
+  case LiftTrueLongDouble:
+    e->lift_longdouble.write(f,e->lift_longdouble.data);
+    break;
   case NameRequest:
     cmixFatal("Some name didn't get managed\n");
   default:
+    fprintf(stderr,"magic=%x\n",e->common.tag);
     cmixFatal("bad expression magic in unparse()\n");
   }
 }
@@ -413,6 +418,7 @@ printDecl(struct cmixDecl *d,struct cmixNameRec names[], FILE *of)
 static void
 printIndent(int indent,FILE *f)
 {
+  if (f==NULL) return;
   while(indent--) {
     putc(' ',f);
     putc(' ',f);
@@ -491,7 +497,7 @@ printStmtSeq(union cmixStmt *s,int indent,struct cmixNameRec names[], FILE *f)
       printIndent(indent,f);
       fprintf(f,"abort();\n");
       printIndent(indent,f);
-      fprintf(f,"/* residual deref of spectime NULL pointer:\n");
+      fprintf(f,"/* residual deref of stray spectime pointer at `{0}':\n");
       printIndent(indent,f);
       fprintf(f," * ");
       printInner(&s->plain.expr,names,f);
@@ -520,340 +526,431 @@ printStmts(union cmixStmt *s,struct cmixNameRec names[],FILE *f)
     }
 }
 
-static void
-printPlain( Interval *interval, int indent,
-            struct cmixNameRec names[], FILE *f)
+/*****************************************************************************/
+
+static void fprintf2( FILE *f, const char *fmt, ... )
 {
-  union cmixStmt *stmt;
-  
-  stmt = (union cmixStmt *)interval->member.basic.label;
-  do {
-    stmt = stmt->common.next;
-    if ( stmt->common.tag == Plain ) {
-      printIndent(indent,f);
-      printInner(&stmt->plain.expr,names,f);
-      fprintf(f,";\n");
-    }
-  } while ( stmt != interval->member.basic.control );
+  va_list pa;
+  if ( f == NULL ) return;
+  va_start(pa,fmt);
+  vfprintf(f,fmt,pa);
+  va_end(pa);
 }
 
-static unsigned
-prepareInterval( Interval *interval, Loop *loop, Interval *ifFollow,
-                 unsigned nextFree )
+static unsigned nextFree;
+
+static Interval *
+ifFollow( Interval *interval )
 {
-  struct basic *info;
-  Interval *latch = NULL;
-  Loop *old_loop = NULL;
-  
-  if ( interval == NULL )
-    return nextFree;
-  
-  if ( loop != NULL )
-    latch = loop->latch;
-  
-  assert( interval->type == basic );
-  
-  info = &interval->member.basic;
-  
-  if ( interval == latch || interval == ifFollow )
-    return nextFree;
-  
-  if ( interval->number != -1 ||
-       ( interval->back_refs != NULL &&
-         interval->back_refs->next != NULL &&
-         loop != info->loop &&
-         ( info->loop == NULL || interval != info->loop->header ) ) ) {
-    if ( loop != NULL ) {
-      /* we're in a loop */
-      if ( ( interval == loop->latch && loop->type == dowile ) ||
-           ( interval == loop->header &&
-             loop->type != dowile ) )
-        return nextFree;
-      else if ( interval == loop->follow )
-        return nextFree;
-    }
-    if ( interval->number == 0 || interval->number == -1 )
-      interval->number = nextFree++;
-    return nextFree;
-  }
-  
-  interval->number = 0;
-  if ( info->loop != NULL && info->loop->header == interval ) {
-    switch ( info->loop->type ) {
-    case endless:
-      assert( interval != info->loop->latch );
-      break;
-      
-    case wile:
-      assert( info->control->common.tag == If );
-      if ( info->loop->latch->number == -1 )
-        info->loop->latch->number = 0;
-      nextFree = prepareInterval(interval->forw_refs->next->interval,
-                                 info->loop,ifFollow,nextFree);
-      return prepareInterval(info->loop->follow,loop,ifFollow, nextFree);
-      
-    case dowile:
-      if ( interval == info->loop->latch ) {
-        if ( info->loop->latch->number == -1 )
-          info->loop->latch->number = 0;
-        return prepareInterval(info->loop->follow,loop,ifFollow,nextFree);
-      }
-      break;
-      
-    default:
-      assert( 0 == 1 );
-    }
-    old_loop = loop;
-    loop = info->loop;
-    latch = loop->latch;
-   
-  }
-  
-  if ( interval->forw_refs == NULL )
-    return nextFree;
-  
-  switch ( info->control->common.tag ) {
-  case If:
-    assert(interval->forw_refs->next->interval != info->ifFollow);
-    nextFree = prepareInterval(interval->forw_refs->next->interval,loop,
-                               info->ifFollow,nextFree);
-    if ( interval->forw_refs->interval != latch &&
-         interval->forw_refs->interval != info->ifFollow )
-      nextFree = prepareInterval(interval->forw_refs->interval,
-                                 loop,info->ifFollow,nextFree);
-    nextFree = prepareInterval(info->ifFollow,loop,ifFollow,nextFree);
-    break;
-    
-  case Goto:
-    nextFree = prepareInterval(interval->forw_refs->interval,loop,
-                               ifFollow,nextFree);
-    break;
-    
-  default:
-    fprintf(stderr,"Tag = %d\n", info->control->common.tag);
-    assert( 0 == 1 );
-  }
-  
-  if ( info->loop != NULL && info->loop->header == interval ) {
-    switch ( info->loop->type ) {
-    case endless:
-      if ( info->loop->latch->number == -1 )
-        info->loop->latch->number = 0;
-      break;
+  return interval->member.basic.ifFollow;
+}
 
-    case dowile:
-      if ( info->loop->latch->number == -1 )
-        info->loop->latch->number = 0;
-      break;
+static Interval *
+thenBranch( Interval *interval )
+{
+  return interval->forw_refs->next->interval;
+}
 
-    default:
-      assert( 0 == 1 );
-    }
+static Interval *
+elseBranch( Interval *interval )
+{
+  return interval->forw_refs->interval;
+}
 
-    nextFree = prepareInterval(info->loop->follow,old_loop,ifFollow,nextFree);
-  }
-  
-  return nextFree;
+static Loop *
+getLoop( Interval *interval )
+{
+  return interval->member.basic.loop;
+}
+
+static Interval *
+loopFollow( Interval *interval )
+{
+  return interval->member.basic.loop->follow;
+}
+
+static Interval *
+loopLatch( Interval *interval )
+{
+  return interval->member.basic.loop->latch;
+}
+
+static int
+printed( Interval *interval )
+{
+  return interval->member.basic.printed;
+}
+
+static Control *
+addLoop( Loop *l, Control *list )
+{
+  Control *c = (Control*)malloc(sizeof(Control));
+
+  assert( loop != NULL );
+
+  c->type = loop;
+  c->info.loop = l;
+  c->next = list;
+
+  return c;
+}
+
+static Control *
+addConditional( Interval *iffollow, Control *list )
+{
+  Control *c = (Control*)malloc(sizeof(Control));
+
+  assert( ifFollow != NULL );
+
+  c->type = conditional;
+  c->info.ifFollow = iffollow;
+  c->next = list;
+
+  return c;
 }
 
 static void
-printInterval( Interval *interval, int indent, struct cmixNameRec names[],
-               Loop *loop, Interval *ifFollow, FILE *f )
+printBB( FILE *f, int indent, struct cmixNameRec names[], Interval *interval )
 {
   union cmixStmt *stmt;
-  struct basic *info;
-  Interval *latch = NULL;
-  Loop *old_loop = NULL;
-  
-  if ( interval == NULL )
-    return;
-  
-  if ( loop != NULL )
-    latch = loop->latch;
-  
-  assert( interval->type == basic );
-  
-  info = &interval->member.basic;
-  
-  if ( interval == latch || interval == ifFollow ) {
-    return;
+
+  assert( ! printed( interval ) );
+  interval->member.basic.printed = 1;
+
+  if ( interval->number > 0 )
+    fprintf2( f, "L%u:\n", interval->number );
+
+  for ( stmt = interval->member.basic.label->next;
+	stmt->common.tag == Plain;
+	stmt = stmt->common.next ) {
+    printIndent( indent, f );
+    printInner(&stmt->plain.expr,names,f);
+    fprintf2(f,";\n");
   }
 
-  
-  if ( info->printed ||
-       ( interval->back_refs != NULL &&
-         interval->back_refs->next != NULL &&
-         loop != info->loop &&
-         ( info->loop == NULL || interval != info->loop->header ) ) ) {
+  assert( stmt == interval->member.basic.control );
+
+  switch ( stmt->common.tag ) {
+  case Return:
     printIndent(indent,f);
-    if ( loop != NULL ) {
-      /* we're in a loop */
-      if ( ( interval == loop->latch && loop->type == dowile ) ||
-           ( interval == loop->header &&
-             loop->type != dowile ) ) {
-        fprintf(f,"continue;\n");
-        return;
-      } else if ( interval == loop->follow ) {
-        fprintf(f,"break;\n" );
-        return;
-      }
-    }
-
-    fprintf(f,"goto L%u;\n",interval->number);
-    
-    assert( interval->number != -1 );
-    assert( interval->number != 0 );
-    
-    return;
-  }
-  
-  info->printed = 1;
-  
-  stmt = (union cmixStmt *)info->label;
-  if ( interval->number != 0 && interval->number != -1 )
-    fprintf(f,"L%u:\n",interval->number);
-  
-  if ( info->loop != NULL && info->loop->header == interval ) {
-    switch ( info->loop->type ) {
-    case endless:
-      assert( interval != info->loop->latch );
-      printIndent(indent,f);
-      fprintf(f,"for ( ; ; ) {\n");
-      break;
-
-    case wile:
-      assert( info->control->common.tag == If );
-      printIndent(indent,f);
-      fprintf(f,"while (");
-      printInner(info->control->cond.cond,names,f);
-      fprintf(f,") {\n");
-      printInterval(interval->forw_refs->next->interval,indent+1,names,
-                    info->loop,ifFollow,f);
-      if ( info->loop->latch->number != 0 &&
-           info->loop->latch->number != -1 )
-        fprintf(f,"L%u:\n",info->loop->latch->number);
-      info->loop->latch->member.basic.printed = 1;
-      printPlain(info->loop->latch,indent+1,names,f);
-      printIndent(indent,f);
-      fprintf(f,"}\n");
-      printInterval(info->loop->follow,indent,names,loop,ifFollow,f);
-      return;
-      
-    case dowile:
-      printIndent(indent,f);
-      fprintf(f,"do {\n");
-      if ( interval == info->loop->latch ) {
-        if ( info->loop->latch->number != 0 &&
-             info->loop->latch->number != -1 )
-          fprintf(f,"L%u:\n",info->loop->latch->number);
-        info->loop->latch->member.basic.printed = 1;
-        printPlain(info->loop->latch,indent+1,names,f);
-        printIndent(indent,f);
-        fprintf(f,"} while (");
-        printInner(info->loop->latch->member.basic.control->cond.cond,
-                   names,f);
-        fprintf(f,");\n");
-        printInterval(info->loop->follow,indent,names,loop,ifFollow,f);
-        return;
-      }
-      break;
-      
-    default:
-      assert( 0 == 1 );
-    }
-    
-    old_loop = loop;
-    loop = info->loop;
-    latch = loop->latch;
-    indent++;
-  }
-  
-  printPlain(interval,indent,names,f);
-  
-  if ( interval->forw_refs == NULL ) {
-    switch ( info->control->common.tag ) {
-    case Return:
-      printIndent(indent,f);
-      fprintf(f,"return ");
-      printInner(&info->control->plain.expr,names,f);
-      fprintf(f,";\n");
-      break;
-    default:
-      assert( 0 == 1 );
-    }
-    return;
-  }
-  
-  /* Not the header of a loop, maybe a structured conditional? */
-  
-  switch ( info->control->common.tag ) {
-  case If:
-    assert(interval->forw_refs->next->interval != info->ifFollow);
+    fprintf2(f,"return ");
+    printInner(&stmt->plain.expr,names,f);
+    fprintf2(f,";\n");
+    break;
+  case Abort:
     printIndent(indent,f);
-    fprintf(f,"if (");
-    printInner(info->control->cond.cond,names,f);
-    fprintf(f,") {\n");
-    printInterval(interval->forw_refs->next->interval,indent+1,names,loop,
-                  info->ifFollow,f);
-    if ( interval->forw_refs->interval != latch &&
-         interval->forw_refs->interval != info->ifFollow ) {
-      printIndent(indent,f);
-      fprintf(f,"} else {\n");
-      printInterval(interval->forw_refs->interval,indent+1,names,loop,
-                    info->ifFollow,f);
-      printIndent(indent,f);
-      fprintf(f,"}\n");
-    } else {
-      printIndent(indent,f);
-      fprintf(f,"}\n");
-    }
-    
-    printInterval(info->ifFollow,indent,names,loop,ifFollow,f);
+    fprintf2(f,"abort();\n");
+    printIndent(indent,f);
+    fprintf2(f,"/* residual deref of stray spectime pointer at `{0}':\n");
+    printIndent(indent,f);
+    fprintf2(f," * ");
+    printInner(&stmt->plain.expr,names,f);
+    fprintf2(f," */\n");
     break;
-    
-  case Goto:
-    printInterval(interval->forw_refs->interval,indent,names,loop,
-                  ifFollow,f);
-    break;
-    
   default:
-    assert( 0 == 1 );
-  }
-  
-  if ( info->loop != NULL && info->loop->header == interval ) {
-    indent--;
-    switch ( info->loop->type ) {
-    case endless:
-      if ( info->loop->latch->number != 0 &&
-           info->loop->latch->number != -1 )
-        fprintf(f,"L%u:\n",info->loop->latch->number);
-      info->loop->latch->member.basic.printed = 1;
-      printPlain(info->loop->latch,indent+1,names,f);
-      printIndent(indent,f);
-      fprintf(f,"}\n");
-      break;
-
-    case dowile:
-      if ( info->loop->latch->number != 0 &&
-           info->loop->latch->number != -1 )
-        fprintf(f,"L%u:\n",info->loop->latch->number);
-      info->loop->latch->member.basic.printed = 1;
-      printPlain(info->loop->latch,indent+1,names,f);
-      printIndent(indent,f);
-      fprintf(f,"} while (");
-      printInner(info->loop->latch->member.basic.control->cond.cond,
-                 names,f);
-      fprintf(f,");\n");
-      break;
-
-    default:
-      assert( 0 == 1 );
-    }
-
-    printInterval(info->loop->follow,indent,names,old_loop,ifFollow,f);
   }
 }
 
+static int
+trivialJump( Interval *target, Control *control )
+{
+  if ( control != NULL ) {
+    switch ( control->type ) {
+    case conditional:
+      /* Falling out of a structured conditional moves you to
+	 the follow. */
+      if ( target == control->info.ifFollow )
+	return 1;
+      break;
+    case loop:
+      /* So we reached the latch of the loop.  This will be
+	 printed automatically immediately after. */
+      if ( target == control->info.loop->latch )
+	return 1;
+      break;
+    default:
+      cmixFatal( "Unknown control structure in trivialJump" );
+    }
+  }
+
+  return 0;
+}
+
+static void
+printJump( FILE *f, int indent, Interval *target, Control *control )
+{
+  Control *cref;
+
+  if ( trivialJump( target, control ) )
+    return;
+
+  printIndent( indent, f );
+  
+  /* Check if a break or continue is applicable */
+  for ( cref = control; cref != NULL; cref = cref->next )
+    if ( cref->type == loop ) {
+      if ( target == cref->info.loop->follow ) {
+	fprintf2( f, "break;\n" );
+	return;
+      } else
+	switch ( cref->info.loop->type ) {
+	case wile:
+	case endless:
+	  if ( target == cref->info.loop->header ) {
+	    fprintf2( f, "continue;\n" );
+	    return;
+	  }
+	  break;
+      case dowile:
+	if ( target == cref->info.loop->latch ) {
+	  fprintf2( f, "continue;\n" );
+	  return;
+	}
+	break;
+	default:
+	  cmixFatal( "Unknown loop type in printJump" );
+	}
+      break;
+    }
+  
+  if ( f == NULL && target->number == 0 )
+    target->number = nextFree++;
+  
+  assert( target->number > 0 );
+  fprintf2( f, "goto L%u;\n", target->number );
+}
+
+static void printInterval( FILE *, int, struct cmixNameRec[], Interval *,
+			   Control * );
+
+static void
+printPlain( FILE *f, int indent, struct cmixNameRec names[],
+	    Interval *interval, Control *control )
+{
+  printBB( f, indent, names, interval );
+  if ( interval->forw_refs != NULL ) {
+    assert( interval->forw_refs->next == NULL );
+    printInterval( f, indent, names, interval->forw_refs->interval, control );
+  }
+}
+
+static void
+printConditional( FILE *f, int indent, struct cmixNameRec names[],
+		  Interval *interval, Control *control )
+{
+  struct basic *info = &interval->member.basic;
+
+  assert( info->control->common.tag == If );
+
+  printBB( f, indent, names, interval );
+
+  printIndent( indent, f );
+  fprintf2( f, "if (" );
+  printInner( info->control->cond.cond, names, f );
+  fprintf2( f, ") {\n" );
+
+  if ( ifFollow( interval ) != NULL ) {
+    Control *newControl = addConditional( info->ifFollow, control );
+
+    printInterval( f, indent + 1, names, thenBranch( interval ), newControl );
+
+    if ( elseBranch( interval ) != ifFollow( interval ) ) {
+      printIndent( indent, f );
+      fprintf2( f, "} else {\n" );
+      printInterval( f, indent + 1, names, elseBranch( interval ),
+		     newControl );
+    }
+
+    free( newControl );
+
+    printIndent( indent, f );
+    fprintf2( f, "}\n" );
+    
+    printInterval( f, indent, names, ifFollow( interval ), control );
+  } else {
+    printInterval( f, indent + 1, names, thenBranch( interval ), control );
+    printIndent( indent, f );
+    fprintf2( f, "} else {\n" );
+    printInterval( f, indent + 1, names, elseBranch( interval ), control );
+    printIndent( indent, f );
+    fprintf2( f, "}\n" );
+  }
+}
+
+static void
+printLoop( FILE *f, int indent, struct cmixNameRec names[],
+	   Interval *interval, Control *control )
+{
+  struct basic *info = &interval->member.basic;
+  Control *newControl = addLoop( info->loop, control );
+
+  switch ( info->loop->type ) {
+  case wile:
+    assert( info->control->common.tag == If );
+    printBB( f, indent + 1, names, interval );
+    printIndent( indent, f );
+    fprintf2( f, "while (" );
+    printInner( info->control->cond.cond, names, f );
+    fprintf2(f,") {\n");
+    break;
+  case dowile:
+    assert( info->loop->latch->member.basic.control->common.tag == If );
+    printIndent( indent, f );
+    fprintf2( f, "do {\n" );
+    if ( info->control->common.tag == If && loopLatch( interval ) != interval )
+      printConditional( f, indent + 1, names, interval, newControl );
+    else
+      printBB( f, indent + 1, names, interval );
+    break;
+  case endless:
+    printIndent( indent, f );
+    fprintf2( f, "for (;;) {\n" );
+    if ( info->control->common.tag == If )
+      printConditional( f, indent + 1, names, interval, newControl );
+    else
+      printBB( f, indent + 1, names, interval );
+    break;
+  default:
+    cmixFatal( "Unknown loop type in printLoop" );
+  }
+
+  if ( interval != loopLatch( interval ) ) {
+
+    /* Now print the body of the loop */
+
+    if ( info->loop->type == wile )
+      printInterval( f, indent + 1, names, interval->forw_refs->next->interval,
+		     newControl );
+    else if ( info->control->common.tag != If )
+      printInterval( f, indent + 1, names, interval->forw_refs->interval,
+		     newControl );
+
+    /* Any latch can only belong to one loop. */
+
+    if ( loopLatch( interval )->member.basic.control->common.tag == If &&
+	 info->loop->type != dowile )
+      printConditional( f, indent + 1, names, loopLatch( interval ),
+			newControl );
+    else
+      printBB( f, indent + 1, names, loopLatch( interval ) );
+  }
+
+  free( newControl );
+
+  /* At this point, the latch node has already been printed.  Either because
+     it was also the header node, or explicitly above. */
+
+  assert( printed( loopLatch( interval ) ) );
+
+  switch ( info->loop->type ) {
+  case wile:
+    printIndent( indent, f );
+    fprintf2( f, "}\n" );
+    break;
+  case dowile:
+    printIndent( indent, f );
+    fprintf2( f, "} while (" );
+    printInner( loopLatch( interval )->member.basic.control->cond.cond,
+		names, f );
+    fprintf2( f, ");\n" );
+    break;
+  case endless:
+    printIndent( indent, f );
+    fprintf2( f, "}\n" );
+    break;
+  default:
+    cmixFatal( "Unknown loop type in printLoop" );
+  }
+
+  if ( loopFollow( interval ) != NULL )
+    printInterval( f, indent, names, loopFollow( interval ), control );
+}
+
+/* printElsewhere decides whether the interval should be printed
+   elsewhere.  The return code is:
+
+   0 - Print here.
+   >0 - print later:
+     1 - Already printed.
+     2 - Latch of loop.
+     3 - Follow of loop or conditional.
+*/
+static int
+printElsewhere( Interval *interval, Control *control )
+{
+  Control *cref;
+
+  /* If the interval has already been printed, then it should obviously
+     not be printed again.  This case also takes care of jumps to
+     sorrounding loops, since the header is (of course) the first
+     interval printed from a loop. */
+
+  if ( printed( interval ) )
+    return 1;
+
+  /* If the interval is the latch of a loop, it will be printed when
+     the loop is printed.  On the other hand, if the interval is also
+     the header of the loop, then this may be the only place to print
+     it. */
+
+  if ( interval->member.basic.loop != NULL &&
+       interval == interval->member.basic.loop->latch &&
+       interval != interval->member.basic.loop->header)
+    return 2;
+
+  /* If the interval is the follow of any structure, we should wait till
+     we're out of the structure before printing it. */
+
+  for ( cref = control; cref != NULL; cref = cref->next )
+    if ( ( cref->type == loop && cref->info.loop->follow == interval ) ||
+	 ( cref->type == conditional && cref->info.ifFollow == interval ) )
+      return 3;
+  
+  return 0;
+}
+
+static void
+printInterval( FILE *f, int indent, struct cmixNameRec names[],
+	       Interval *interval, Control *control )
+{
+  struct basic *info;
+
+  assert( interval != NULL );
+  assert( interval->type == basic );
+
+  info = &interval->member.basic;
+
+  /* If this interval will be or is printed elsewhere, print a jump. */
+  if ( printElsewhere( interval, control ) ) {
+    printJump( f, indent, interval, control );
+    return;
+  }
+  
+  if ( getLoop( interval ) != NULL && interval == info->loop->header )
+    printLoop( f, indent, names, interval, control );
+  else if ( info->control->common.tag == If )
+    printConditional( f, indent, names, interval, control );
+  else
+    printPlain( f, indent, names, interval, control );
+}
+
+static void
+resetLabels( unsigned count, Interval *order[] )
+{
+  int idx;
+
+  for ( idx = 1; idx <= count; idx++ )
+      order[idx]->number = 0;
+}
+
+static void
+resetPrint( unsigned count, Interval *order[] )
+{
+  int idx;
+
+  for ( idx = 1; idx <= count; idx++ )
+    order[idx]->member.basic.printed = 0;
+}
 
 /*****************************************************************************/
 
@@ -947,37 +1044,12 @@ cmixUnparse(struct cmixNameRec names[], unsigned const namesSorted[],
             fprintf(outfile,";\n");
           }
         if ( cmixRestruct ) {
-          Restruct *res;
-          int printSep = 0;
-          int idx;
-          int nextFree = 1;
-          res = cmixRestructStmts(f->stmts);
-          do {
-            for ( idx = 1; idx <= res->count; idx++ ) {
-              if ( res->order[idx]->number == -1 ) {
-                nextFree = prepareInterval(res->order[idx],NULL,NULL,nextFree);
-                continue;
-              }
-            }
-            break;
-          } while ( 1 );
-          
-          do {
-            for ( idx = 1; idx <= res->count; idx++ ) {
-              if ( ! res->order[idx]->member.basic.printed ) {
-                if ( printSep ) {
-                  fprintf(outfile,"\n/******************/\n");
-                  fprintf(outfile,  "/** END OF BLOCK **/\n");
-                  fprintf(outfile,  "/******************/\n\n");
-                }
-                printSep = 1;
-                
-                printInterval(res->order[idx],1,names,NULL,NULL,outfile);
-                continue;
-              }
-            }
-            break;
-          } while ( 1 );
+          Restruct *res = cmixRestructStmts(f->stmts);
+	  nextFree = 1;
+	  resetLabels( res->count, res->order );
+	  printInterval( NULL, 1, names, res->order[1], NULL );
+	  resetPrint( res->count, res->order );
+	  printInterval( outfile, 1, names, res->order[1], NULL );
         } else {
           printStmts(f->stmts,names,outfile);
         }
